/***************************************************************************
 *   Copyright (C) 2007 kommute dev team                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "fileShare.h"
#include "MUTE/layers/messageRouting/messageRouter.h"
#include "MUTE/common/CryptoUtils.h"


#include <QFile>
#include "minorGems/io/file/File.h"
#include "minorGems/io/file/Path.h"
#include "minorGems/util/stringUtils.h"
#include "minorGems/util/SettingsManager.h"
#include "minorGems/network/web/URLUtils.h"
#include "minorGems/network/web/MimeTyper.h"
#include "minorGems/formats/encodingUtils.h"

#include <QSemaphore>
#include <QMutex>
#include <QTime>
#include <QThread>
#include "MUTE/FinishedSignalThread.h"
#include "MUTE/StopSignalThread.h"

#include "minorGems/util/SimpleVector.h"

#include "minorGems/util/log/AppLog.h"

#include "MUTE/sha1.h"
#include <QCryptographicHash>


#include "minorGems/util/random/RandomSource.h"
#include "minorGems/util/random/StdRandomSource.h"

#include <QDateTime>
#include "settings.h"
extern class Settings *settings;

typedef struct sharehash
{
  char *name;
  char *hash;
  qint64 size;
  QDateTime *time;
} SHASH;

extern SHASH *Shares;
extern int maxShares;
extern int nbShares;


int muteShareFileInfoSenderID = -1;
int muteShareFileChunkSenderID = -1;
int muteShareSearchResultsSenderID = -1;
MimeTyper *muteShareMimeTyper = NULL;
char *muteShareVirtualAddress = NULL;

QMutex muteUploadDataLock;
int muteNextUploadID = 0;
SimpleVector < long >*muteUploadIDs = NULL;
SimpleVector < char *>*muteUploadFilePaths = NULL;
SimpleVector < char *>*muteUploadHostAddresses = NULL;
SimpleVector < long >*muteUploadLastChunksSent = NULL;
SimpleVector < long >*muteUploadChunksInFiles = NULL;
SimpleVector < unsigned long >*muteUploadFirstChunkTimes = NULL;
SimpleVector < unsigned long >*muteUploadLastChunkTimes = NULL;

class SourceQueue;

QMutex muteDownloadDataLock;
int muteNextDownloadID = 0;
SimpleVector < long >*muteDownloadIDs = NULL;
SimpleVector < QSemaphore * >*muteDownloadSemaphores = NULL;
SimpleVector < char >*muteDownloadInterruptedFlags = NULL;
SimpleVector < SourceQueue * >*muteDownloadSourceQueues = NULL;

long muteNumActiveDownloads = 0;
long muteMaxSimultaneousDownloads = 0;
// queue of download IDs that are waiting for a slot
SimpleVector < long >*muteDownloadQueue = NULL;


// chunk size is 16 KiB.  After base64 encoding, it will take up less than
// 22 KiB in a MUTE message body, leaving room for various headers.
// The size limit on a MUTE message is 32 KiB
//
// In previous versions, the chunk size was 22 KiB before encoding.
// We now use 16KiB so that the chunk size is a power of 2 and therefore
// compatible with many partial hashing schemes (TigerTrees, in particular)
int muteShareChunkSize = 16384;


char muteSharePrintSearchSyncTrace = false;




// tag that is present in all internal directories that should not
// be remotely searchable
char *muteInternalDirectoryTag = "MUTE_INTERNAL";

// sub-directory in shared directory where file hashes are cached 
//char *muteHashDirectoryName = "MUTE_INTERNAL_hashes";
// sub-directory in shared directory where incoming files are kept 
char *muteIncomingDirectoryName = "MUTE_INTERNAL_incoming";



/**
 * Translates a URL-safe encoded path into a file object.
 *
 * @param inEncodedPath the URL-safe path.  Must be destroyed by caller.
 *
 * @return a file object, or NULL if the file does not exist.
 *   Must be destroyed by caller.
 */
File *
muteShare_internalEncodedPathToFile (char *inEncodedPath)
{
  char *path = URLUtils::hexDecode (inEncodedPath);

  if (strstr (path, ".."))
    {
      // path may be trying to escape from our file directory
      // block it

      // note:  this may also block file names with "..", like,
      // test..txt
      // but who really names files like this anyway?
      // should really look for /.. or ../, but this might not be as
      // safe (what if, on windows, someone requests \.. directly,
      //       even though this isn't a valid MUTE path)

      delete[]path;
      return NULL;
    }

  char *pathPointer = path;

  if (pathPointer[strlen (pathPointer) - 1] == '/')
    {
      // path ends with /, remove it
      pathPointer[strlen (pathPointer) - 1] = '\0';
    }
  if (pathPointer[0] == '/')
    {
      // path starts with /, skip it
      pathPointer = &(pathPointer[1]);
    }


  char *platformPath;

  // transform it into a platform-specific path
  char platformDelim = Path::getDelimeter ();

  if (platformDelim == '/')
    {
      // already in platform-specific form
      platformPath = stringDuplicate (pathPointer);
    }
  else
    {
      char found;

      char *platformDelimString = autoSprintf ("%c", platformDelim);

      // replace protocol delimeters with platform-specific delimiters
      platformPath = replaceAll (pathPointer, "/",
				 platformDelimString, &found);

      delete[]platformDelimString;
    }

  delete[]path;


  char *sharingPathString = muteShareGetSharingPath ();

  File *file;

  if (strlen (platformPath) == 0)
    {
      // root directory
      file = new File (NULL, sharingPathString);
    }
  else
    {
      Path *sharingPath = new Path (sharingPathString);
      file = new File (sharingPath, platformPath);
    }
  delete[]sharingPathString;

  delete[]platformPath;

  if (file->exists ())
    {
      return file;
    }
  else
    {
      AppLog::error ("fileTransfer", "File does not exist");
      delete file;
      return NULL;
    }
}



/**
 * Translates a file object into a URL-safe encoded path.
 *
 * Note that this call returns NULL for files that are inside
 * MUTE_INTERNAL_hashes and MUTE_INTERNAL_incoming directories.
 *
 * @param inFile a file object.  Must be destroyed by caller.
 *
 * @return the URL-safe path, or NULL on error.
 *   Must be destroyed by caller.
 */
char *
muteShare_internalFileToEncodedPath (File * inFile)
{

  char platformDelim = Path::getDelimeter ();

  char *sharingPathString = muteShareGetSharingPath ();

  char *fullFilePathString = inFile->getFullFileName ();


  if (strstr (fullFilePathString, muteInternalDirectoryTag) != NULL)
    {

      // ignore files in our internal directories

      delete[]sharingPathString;
      delete[]fullFilePathString;

      return NULL;
    }


  // remove sharing path from file name
  char *pointerToSharingPath = strstr (fullFilePathString,
				       sharingPathString);

  char *returnPathString = NULL;
  if (pointerToSharingPath != NULL)
    {

      // skip the sharing path
      char *partialFilePath =
	&(fullFilePathString[strlen (sharingPathString)]);


      if (strlen (partialFilePath) > 0)
	{
	  // make sure we didn't miss the final delimiter
	  if (partialFilePath[0] == platformDelim)
	    {

	      // skip the first character (delimeter
	      partialFilePath = &(partialFilePath[1]);
	    }
	}

      if (platformDelim == '/')
	{
	  // already using universal path delimeter
	  returnPathString = stringDuplicate (partialFilePath);
	}
      else
	{
	  // replace the platform-specific path delimeter with the universal
	  // delimeter "/"
	  char *platformDelimString = autoSprintf ("%c", platformDelim);

	  char found;

	  returnPathString = replaceAll (partialFilePath,
					 platformDelimString, "/", &found);

	  delete[]platformDelimString;
	}
    }

  delete[]sharingPathString;
  delete[]fullFilePathString;

  return returnPathString;
}



// handler for FileInfoRequests
int
muteShare_internalFileInfoSender (char *inFromAddress, char *inToAddress,
				  char *inBody, void *inExtraArgument)
{

  // is the message a FileInfoRequest?
  // look at second token
  SimpleVector < char *>*tokens = tokenizeString (inBody);
  int numTokens = tokens->size ();

  if (numTokens == 4)
    {
      char *typeToken = *(tokens->getElement (1));
      if (strcmp ("FileInfoRequest", typeToken) == 0)
	{

	  char *encodedPath = *(tokens->getElement (3));

	  File *file = muteShare_internalEncodedPathToFile (encodedPath);

	  if (file != NULL && !file->isDirectory ())
	    {

	      // file exists, send an info message
	      int fileLength;
	      char *mimeType;
	      int numChunks;


	      fileLength = file->getLength ();

	      char *fileName = file->getFileName ();

	      mimeType = muteShareMimeTyper->getFileNameMimeType (fileName);

	      delete[]fileName;

	      if (mimeType == NULL)
		{
		  // default type
		  mimeType = stringDuplicate ("application/octet-stream");
		}

	      numChunks = fileLength / muteShareChunkSize;

	      if (fileLength % muteShareChunkSize != 0)
		{
		  // extra partial chunk
		  numChunks += 1;
		}

	      char *message = autoSprintf ("MessageType: FileInfo\n"
					   "FilePath: %s\n"
					   "FileStatus: Found\n"
					   "FileSize: %d\n"
					   "ChunkCount: %d\n" "MimeType: %s",
					   encodedPath,
					   fileLength,
					   numChunks,
					   mimeType);

	      delete[]mimeType;

	      // use ROUTE_ONLY flag when sending back FileInfo
	      muteSendMessage (muteShareVirtualAddress,
			       inFromAddress, message, "ROUTE_ONLY");

	      delete[]message;
	    }
	  else
	    {
	      AppLog::error ("fileTransfer",
			     "FileInfoRequest does not contain a valid file path");

	      // send back a not found message
	      char *message = autoSprintf ("MessageType: FileInfo\n"
					   "FilePath: %s\n"
					   "FileStatus: NotFound",
					   encodedPath);

	      // use ROUTE_ONLY flag when sending back FileInfo
	      muteSendMessage (muteShareVirtualAddress,
			       inFromAddress, message, "ROUTE_ONLY");

	      delete[]message;
	    }

	  if (file != NULL)
	    {
	      delete file;
	    }

	}

    }

  for (int i = 0; i < numTokens; i++)
    {
      delete[] * (tokens->getElement (i));
    }

  delete tokens;

  // no utility generated
  return 0;
}



// reports that a chunk has been sent for a file.
// inFilePath and inHostAddress must be destroyed by caller
void
muteShare_internalChunkSent (char *inFilePath,
			     char *inHostAddress,
			     int inChunkNumber, int inChunksInFile)
{
  // get current time in seconds
  unsigned long currentSeconds = time (NULL);

  // Search for the file in our upload data; both file path and destination
  // address must match to be considered the same entry.  Multiple uploads
  // of the same file to different destinations are recorded separately.
  char found = false;
  int foundIndex = -1;

  muteUploadDataLock.lock ();

  int numUploads = muteUploadIDs->size ();
  for (int i = 0; i < numUploads && !found; i++)
    {

      if (strcmp (inFilePath,
		  *(muteUploadFilePaths->getElement (i))) == 0 &&
	  strcmp (inHostAddress,
		  *(muteUploadHostAddresses->getElement (i))) == 0)
	{

	  // both file path and host address match

	  found = true;
	  foundIndex = i;
	}
    }

  if (found)
    {
      // save last chunk sent index
      long *indexPointer = muteUploadLastChunksSent->getElement (foundIndex);
      *indexPointer = inChunkNumber;

      // save time of last chunk send
      unsigned long *timePointer
	= muteUploadLastChunkTimes->getElement (foundIndex);
      *timePointer = currentSeconds;
    }
  else
    {
      // add a new upload to end of vector
      muteUploadIDs->push_back (muteNextUploadID);
      muteNextUploadID++;

      muteUploadFilePaths->push_back (stringDuplicate (inFilePath));
      muteUploadHostAddresses->push_back (stringDuplicate (inHostAddress));
      muteUploadLastChunksSent->push_back (inChunkNumber);
      muteUploadChunksInFiles->push_back (inChunksInFile);
      muteUploadFirstChunkTimes->push_back (currentSeconds);
      muteUploadLastChunkTimes->push_back (currentSeconds);
    }

  muteUploadDataLock.unlock ();
}



// handler for FileChunkRequests
int
muteShare_internalFileChunkSender (char *inFromAddress, char *inToAddress,
				   char *inBody, void *inExtraArgument)
{
  // is the message a FileChunkRequest?
  // look at second token
  SimpleVector < char *>*tokens = tokenizeString (inBody);
  int numTokens = tokens->size ();

  if (numTokens == 6)
    {
      char *typeToken = *(tokens->getElement (1));
      if (strcmp ("FileChunkRequest", typeToken) == 0)
	{

	  char *encodedPath = *(tokens->getElement (3));

	  File *file = muteShare_internalEncodedPathToFile (encodedPath);

	  if (file != NULL)
	    {

	      // make sure they're not requesting chunks of a directory
	      // file 
	      if (!file->isDirectory ())
		{

		  unsigned long fileLength = file->getLength ();

		  int chunksInFile = fileLength / muteShareChunkSize;

		  if (fileLength % muteShareChunkSize != 0)
		    {
		      // extra partial chunk
		      chunksInFile += 1;
		    }

		  char *chunkNumberString = *(tokens->getElement (5));

		  int chunkNumber;

		  int numRead = sscanf (chunkNumberString, "%d",
					&chunkNumber);

		  if (numRead == 1 && chunkNumber < chunksInFile)
		    {

		      // send a FileChunk message

		      char *fileName = file->getFullFileName ();

		      FILE *fileHandle = fopen (fileName, "rb");



		      if (fileHandle != NULL)
			{
			  unsigned long bytesToSkip =
			    chunkNumber * muteShareChunkSize;

			  fseek (fileHandle, bytesToSkip, SEEK_SET);

			  int chunkSize = muteShareChunkSize;
			  if (bytesToSkip + muteShareChunkSize > fileLength)
			    {

			      // partial chunk
			      chunkSize = fileLength - bytesToSkip;
			    }

			  unsigned char *rawChunkData =
			    new unsigned char[chunkSize];

			  numRead = fread (rawChunkData, 1, chunkSize,
					   fileHandle);

			  if (numRead == chunkSize)
			    {

			      // base64 encode with no line breaks
			      char *encodedChunkData =
				base64Encode (rawChunkData, chunkSize,
					      false);

			      char *message =
				autoSprintf ("MessageType: FileChunk\n"
					     "FilePath: %s\n"
					     "ChunkNumber: %d\n"
					     "ChunkLength: %d\n"
					     "ChunkData: %s",
					     encodedPath,
					     chunkNumber,
					     chunkSize,
					     encodedChunkData);

			      delete[]encodedChunkData;

			      // use ROUTE_ONLY flag to avoid wasting
			      // bandwidth on lost chunks
			      muteSendMessage (muteShareVirtualAddress,
					       inFromAddress,
					       message, "ROUTE_ONLY");
			      delete[]message;

			      muteShare_internalChunkSent (fileName,
							   inFromAddress,
							   chunkNumber,
							   chunksInFile);
			    }

			  delete[]rawChunkData;

			  fclose (fileHandle);
			}

		      delete[]fileName;
		    }
		}
	      delete file;
	    }
	}
    }

  for (int i = 0; i < numTokens; i++)
    {
      delete[] * (tokens->getElement (i));
    }

  delete tokens;

  // no utility generated
  return 0;
}



// utility penalty for searches that contain no terms
// should cause them to be dropped quickly
int muteShareBlankSearchUtilityPenalty = 1000;



// handler for SearchRequests
int
muteShare_internalSearchResultsSender (char *inFromAddress,
				       char *inToAddress,
				       char *inBody, void *inExtraArgument)
{
  int utility = 0;

  // is the message a SearchRequest?
  // look at second token
  SimpleVector < char *>*tokens = tokenizeString (inBody);
  int numTokens = tokens->size ();

  if (numTokens == 6)
    {
      char *typeToken = *(tokens->getElement (1));
      if (strcmp ("SearchRequest", typeToken) == 0)
	{

	  char *searchID = *(tokens->getElement (3));
	  char *encodedSearchString = *(tokens->getElement (5));

	  char *searchString = URLUtils::hexDecode (encodedSearchString);


	  char *logMessage =
	    autoSprintf ("Got search request for (%s), id=%s",
			 searchString, searchID);
	  AppLog::detail ("fileShare", logMessage);
	  delete[]logMessage;


	  // check if this is a hash-only search
	  char hashOnlySearch = false;

	  const char *hashStart = "hash_";

	  char *pointerToHashStart = strstr (searchString, hashStart);
	  if (pointerToHashStart != NULL)
	    {
	      // search contains a hash-only search.
	      // extract hash, and ignore any other search terms

	      char *pointerToHash = &(pointerToHashStart[strlen (hashStart)]);

	      // hash should be at most 40 characters long
	      char *hashString = new char[41];

	      int numRead = sscanf (pointerToHash, "%40s", hashString);

	      if (numRead == 1)
		{
		  delete[]searchString;
		  searchString = hashString;

		  hashOnlySearch = true;
		}
	      else
		{
		  delete[]hashString;
		}
	    }



	  SimpleVector < char *>*searchTerms = tokenizeString (searchString);
	  int numSearchTerms = searchTerms->size ();

	  delete[]searchString;

	  char *sharedPath = muteShareGetSharingPath ();

	  File *sharedDirectory = new File (NULL, sharedPath);
	  delete[]sharedPath;


	  // penalize searches that contain no terms
	  if (numSearchTerms == 0)
	    {
	      utility += muteShareBlankSearchUtilityPenalty;
	    }

	  // ignore searches that contain no terms
	  // only search if we have a valid shared directory
	  if (numSearchTerms > 0 &&
	      sharedDirectory->exists () && sharedDirectory->isDirectory ())
	    {

	      SimpleVector < char *>*hits = new SimpleVector < char *>();

	      // for now, trim our results so that they fit
	      // in one message

	      // save 4 KiB for our headers, 28 KiB for results
	      int maxResultsLength = 28672;
	      int totalResultLength = 0;

	      char hitLimit = false;

	      int i;
	      for (i = 0; i < nbShares && !hitLimit; i++)
		{

		  char *fileName = Shares[i].name;

		  if (fileName != NULL)
		    {

		      char hitAll = true;

		      if (!hashOnlySearch)
			{
			  // check each term
			  for (int j = 0; j < numSearchTerms && hitAll; j++)
			    {

			      char *term = *(searchTerms->getElement (j));
			      if (stringLocateIgnoreCase (fileName,
							  term) == NULL)
				{

				  // missed this term
				  hitAll = false;
				}
			    }
			}

		      if (hitAll)
			{

			  char *hashString = Shares[i].hash;
			  if (hashString == NULL)
			    {
			      hashString = "";
			    }


			  // check hash if this is a hash-only search
			  char resultIsAHit = true;

			  if (hashOnlySearch)
			    {
			      char *hashTerm = *(searchTerms->getElement (0));
			      if (stringCompareIgnoreCase (hashString,
							   hashTerm) != 0)
				{

				  resultIsAHit = false;
				}
			    }


			  if (resultIsAHit)
			    {

			      char *encodedFileName =
				URLUtils::hexEncode (fileName);

			      char *resultString = autoSprintf ("%s %lld %s",
								encodedFileName,
								Shares[i].
								size,
								hashString);
			      delete[]encodedFileName;

			      int resultLength = strlen (resultString);

			      if (totalResultLength +
				  resultLength + 1 < maxResultsLength)
				{
				  // not at limit yet
				  hits->push_back (resultString);
				  totalResultLength += resultLength + 1;
				}
			      else
				{
				  // hit limit
				  hitLimit = true;
				  delete[]resultString;
				}
			    }
			}
		    }
		}


	      int numHits = hits->size ();

	      if (numHits != 0)
		{

		  logMessage =
		    autoSprintf
		    ("Sending response with %d results, id=%s", numHits,
		     searchID);
		  AppLog::detail ("fileShare", logMessage);
		  delete[]logMessage;


		  char **hitArray = hits->getElementArray ();

		  // entries delimited by newlines
		  char *hitString = join (hitArray,
					  numHits,
					  "\n");
		  for (int i = 0; i < numHits; i++)
		    {
		      delete[]hitArray[i];
		    }
		  delete[]hitArray;


		  char *message =
		    autoSprintf ("MessageType: SearchResults\n"
				 "SearchID: %s\n" "ResultCount: %d\n"
				 "Results: %s",
				 searchID,
				 numHits,
				 hitString);
		  delete[]hitString;

		  // use ROUTE_ONLY flag to reduce excess traffic
		  // from lost search results
		  muteSendMessage (muteShareVirtualAddress,
				   inFromAddress, message, "ROUTE_ONLY");

		  // utility is number of results we return
		  utility += numHits;

		  delete[]message;
		}

	      delete hits;
	    }

	  delete sharedDirectory;

	  for (int j = 0; j < numSearchTerms; j++)
	    {
	      char *term = *(searchTerms->getElement (j));
	      delete[]term;
	    }
	  delete searchTerms;
	}
    }

  for (int i = 0; i < numTokens; i++)
    {
      delete[] * (tokens->getElement (i));
    }

  delete tokens;

  return utility;
}



/**
 * Container for information about a source.
 */
class FileSource
{

public:

	/**
         * Constructs a source.
         *
         * Chunk time starts at 0.
         *
         * @param inVirtualAddress the virtual address.
         *   Will be destroyed when this class is destroyed.
         * @param inFilePath the file path.
         *   Will be destroyed when this class is destroyed.
         */
  FileSource (char *inVirtualAddress,
	      char *inFilePath):mVirtualAddress (inVirtualAddress),
    mFilePath (inFilePath), mChunkTimeInMilliseconds (0)
  {

  }



   ~FileSource ()
  {
    delete[]mVirtualAddress;
    delete[]mFilePath;
  }



  char *mVirtualAddress;
  char *mFilePath;
  long mChunkTimeInMilliseconds;
};



class SourceSearchThread;



/**
 * Tracks a list of possible sources for a file.
 *
 * Automatically searches for more sources when too many known sources have
 * failed.
 *
 * All calls are thread-safe.
 *
 * @author Jason Rohrer.
 */
class SourceQueue
{

public:

	/**
         * Constructs a queue with one source in it.
         *
         * Does not automatically search to fill the queue.
         *
         * @param inFileHash the file hash.
         *   Must be destroyed by caller.
         * @param inFirstVirtualAddress the address of the first source
         *   that we know about.
         *   Must be destroyed by caller.
         * @param inFirstFilePath the file path for the first source.
         *   Must be destroyed by caller.
         */
  SourceQueue (char *inFileHash,
	       char *inFirstVirtualAddress, char *inFirstFilePath);


   ~SourceQueue ();



	/**
         * Report the time to receive a chunk from the best source (the last
         * source returned by getNextSource).
         *
         * @param inMilliseconds the time in milliseconds, or -1 if
         *   Fetching the chunk failed.
         */
  void reportBestSourceChunkTime (long inMilliseconds);



	/**
         * Gets the next source from the queue.
         * Since queue is never empty, this call will always return a source.
         *
         * @param outVirtualAddress pointer to where the virtual address
         *   should be returned.
         *   Returned value must be destroyed by caller.
         * @param outFilePath pointer to where the file path
         *   should be returned.
         *   Returned value must be destroyed by caller.
         */
  void getNextSource (char **outVirtualAddress, char **outFilePath);




	/**
         * Performs a search for more sources and updates mLongestSourceList
         * count.
         *
         * The search is launched in a background thread, so this call
         * returns immediately.
         *
         * This call has no effect if a search is already running.
         *
         * @param inSearchTimeoutInMilliseconds how long to wait for
         *   results to come back before ending search.
         *
         * @return true if a new search was started, or false if a search
         *   is already running in the background.
         */
  char searchForMoreSources (unsigned long inSearchTimeoutInMilliseconds);



	/**
         * Same as searchForMoreSources, but blocks until search completes.
         *
         * @param inThread the thread we are running as part of.
         */
  void searchForMoreSources_blocking (SourceSearchThread * inThread);



	/**
         * Adds a source to this queue, or refreshes an existing source as
         * new.
         *
         * @param inSource the source to add.
         *   Must be destroyed by caller.
         */
  void addSource (FileSource * inSource);



	/**
         * Gets the number of active sources (non-failed) in the queue.
         *
         * @return the number of active sources.
         */
  int getActiveSourceCount ();



	/**
         * Sets the hibernation state for this queue.
         *
         * @param inHibernating true if hibernating, or false otherwise.
         */
  void setHibernating (char inHibernating)
  {
    mLock.lock ();
    mHibernating = inHibernating;
    mLock.unlock ();
  }



	/**
         * Gets the hibernation state for this queue.
         *
         * @return true if hibernating, or false otherwise.
         */
  char isHibernating ()
  {
    mLock.lock ();
    char returnValue = mHibernating;
    mLock.unlock ();

    return returnValue;
  }



protected:
  QMutex mLock;

  char mHibernating;

  SourceSearchThread *mLastThread;


  char *mFileHash;
  SimpleVector < FileSource * >*mSourceVector;


  // best count of non-failed hosts we have seen so far
  long mLongestSourceList;


  QTime mLastSearchStart;


	/**
         * Looks in queue for another source to assume position 0.
         */
  void findNewBest ();

};



/**
 * Checks if current address and message signing keypair are still safe
 * to use (checks if replay attacks have become possible).
 *
 * If the address is no longer safe, a new one is generated.
 *
 * If the existing address is NULL, this call can be used to generate our
 * first address upon startup.
 */
void
muteShare_internalMakeAddressSafe ()
{

  char generateNew = false;

  if (muteHasCounterWrappedAround ())
    {
      generateNew = true;
      if (muteShareVirtualAddress != NULL)
	{

	  muteRemoveReceiveAddress (muteShareVirtualAddress);

	  delete[]muteShareVirtualAddress;
	  muteShareVirtualAddress = NULL;
	}
    }
  else if (muteShareVirtualAddress == NULL)
    {
      generateNew = true;
    }


  if (!generateNew)
    {
      return;
    }


  // the messageRouter layer will sign all of our outbound messages
  // we need to generate our keys for this first (our address is the
  // SHA1 hash of our public key)


  char keyLengthFound;
  int keyLength = SettingsManager::getIntSetting ("messageSigningKeySize",
						  &keyLengthFound);

  if (!keyLengthFound)
    {
      // defaults to 512
      keyLength = 512;
    }

  printf ("Generating new %d-bit RSA key pair for signing messages...\n",
	  keyLength);

  char *pubKey;
  char *privKey;

  CryptoUtils::generateRSAKey (keyLength, &pubKey, &privKey);

  printf ("...done\n");

  char *pubKeyHash = computeSHA1Digest (pubKey);

  // our address is PKH followed by the hash of our pubkey
  muteShareVirtualAddress = autoSprintf ("PKH%s", pubKeyHash);

  delete[]pubKeyHash;

  muteSetMessageIDSigningKeys (pubKey, privKey);
  delete[]pubKey;
  delete[]privKey;


  muteAddReceiveAddress (muteShareVirtualAddress);
}



/**
 * Thread that wakes periodically and makes sure our address is still safe.
 */
class AddressSafetyThread:public StopSignalThread
{

public:



	/**
         * Creates and starts a thread.
         */
  AddressSafetyThread ()
  {
    start ();
  }



	/**
         * Stops and destroys a thread.
         */
   ~AddressSafetyThread ()
  {
    stop ();
  }



  // implements the Thread interface
  virtual void run ()
  {

    while (!isStopped ())
      {
	// wake every 10 minutes
	sleep (10 * 60 * 1000);

	if (!isStopped ())
	  {

	    AppLog::info ("AddressSafetyThread",
			  "Checking to make sure address is still safe.");

	    muteShare_internalMakeAddressSafe ();
	  }
      }
  }

};



AddressSafetyThread *muteShareAddressSafetyThread = NULL;



void
muteShareStart ()
{

  muteMaxSimultaneousDownloads = muteShareGetMaxSimultaneousDownloads ();


  char settingFound;
  int traceSearchFlag =
    SettingsManager::getIntSetting ("printSearchSyncTrace",
				    &settingFound);

  if (settingFound && traceSearchFlag == 1)
    {

      muteSharePrintSearchSyncTrace = true;
    }
  else
    {
      // default
      muteSharePrintSearchSyncTrace = false;
    }



  muteUploadIDs = new SimpleVector < long >();
  muteUploadFilePaths = new SimpleVector < char *>();
  muteUploadHostAddresses = new SimpleVector < char *>();
  muteUploadLastChunksSent = new SimpleVector < long >();
  muteUploadChunksInFiles = new SimpleVector < long >();
  muteUploadFirstChunkTimes = new SimpleVector < unsigned long >();
  muteUploadLastChunkTimes = new SimpleVector < unsigned long >();

  muteDownloadIDs = new SimpleVector < long >();
  muteDownloadSemaphores = new SimpleVector < QSemaphore * >();
  muteDownloadInterruptedFlags = new SimpleVector < char >();
  muteDownloadSourceQueues = new SimpleVector < SourceQueue * >();

  muteDownloadQueue = new SimpleVector < long >();

  muteShareMimeTyper = new MimeTyper ();

  muteShareFileInfoSenderID =
    muteAddMessageHandler (muteShare_internalFileInfoSender, (void *) NULL);
  muteShareFileChunkSenderID =
    muteAddMessageHandler (muteShare_internalFileChunkSender, (void *) NULL);
  muteShareSearchResultsSenderID =
    muteAddMessageHandler (muteShare_internalSearchResultsSender,
			   (void *) NULL);


  // receive broadcasts
  muteAddReceiveAddress ("ALL");


  // generate our first message-signing key pair and virtual address
  muteShare_internalMakeAddressSafe ();

  // start thread to keep address safe
  muteShareAddressSafetyThread = new AddressSafetyThread ();
}



void
muteShareStop ()
{

  if (muteShareAddressSafetyThread != NULL)
    {
      delete muteShareAddressSafetyThread;
      muteShareAddressSafetyThread = NULL;
    }

  if (muteShareVirtualAddress != NULL)
    {
      muteRemoveReceiveAddress (muteShareVirtualAddress);
      delete[]muteShareVirtualAddress;
      muteShareVirtualAddress = NULL;
    }
  muteRemoveReceiveAddress ("ALL");

  if (muteShareSearchResultsSenderID != -1)
    {
      muteRemoveMessageHandler (muteShareSearchResultsSenderID);
      muteShareSearchResultsSenderID = -1;
    }
  if (muteShareFileChunkSenderID != -1)
    {
      muteRemoveMessageHandler (muteShareFileChunkSenderID);
      muteShareFileChunkSenderID = -1;
    }
  if (muteShareFileInfoSenderID != -1)
    {
      muteRemoveMessageHandler (muteShareFileInfoSenderID);
      muteShareFileInfoSenderID = -1;
    }

  if (muteShareMimeTyper != NULL)
    {
      delete muteShareMimeTyper;
      muteShareMimeTyper = NULL;
    }

  if (muteUploadIDs != NULL)
    {
      delete muteUploadIDs;
      muteUploadIDs = NULL;
    }
  if (muteUploadFilePaths != NULL)
    {
      int numPaths = muteUploadFilePaths->size ();
      for (int i = 0; i < numPaths; i++)
	{
	  delete[] * (muteUploadFilePaths->getElement (i));
	}
      delete muteUploadFilePaths;
      muteUploadFilePaths = NULL;
    }
  if (muteUploadHostAddresses != NULL)
    {
      int numPaths = muteUploadHostAddresses->size ();
      for (int i = 0; i < numPaths; i++)
	{
	  delete[] * (muteUploadHostAddresses->getElement (i));
	}
      delete muteUploadHostAddresses;
      muteUploadHostAddresses = NULL;
    }
  if (muteUploadLastChunksSent != NULL)
    {
      delete muteUploadLastChunksSent;
      muteUploadLastChunksSent = NULL;
    }
  if (muteUploadChunksInFiles != NULL)
    {
      delete muteUploadChunksInFiles;
      muteUploadChunksInFiles = NULL;
    }
  if (muteUploadFirstChunkTimes != NULL)
    {
      delete muteUploadFirstChunkTimes;
      muteUploadFirstChunkTimes = NULL;
    }
  if (muteUploadLastChunkTimes != NULL)
    {
      delete muteUploadLastChunkTimes;
      muteUploadLastChunkTimes = NULL;
    }
  if (muteDownloadIDs != NULL)
    {
      delete muteDownloadIDs;
      muteDownloadIDs = NULL;
    }
  if (muteDownloadSemaphores != NULL)
    {
      int numSems = muteDownloadSemaphores->size ();
      for (int i = 0; i < numSems; i++)
	{
	  delete *(muteDownloadSemaphores->getElement (i));
	}
      delete muteDownloadSemaphores;
      muteDownloadSemaphores = NULL;
    }
  if (muteDownloadInterruptedFlags != NULL)
    {
      delete muteDownloadInterruptedFlags;
      muteDownloadInterruptedFlags = NULL;
    }
  if (muteDownloadSourceQueues != NULL)
    {
      int numQueues = muteDownloadSourceQueues->size ();
      for (int i = 0; i < numQueues; i++)
	{
	  delete *(muteDownloadSourceQueues->getElement (i));
	}
      delete muteDownloadSourceQueues;
      muteDownloadSourceQueues = NULL;
    }

  if (muteDownloadQueue != NULL)
    {
      delete muteDownloadQueue;
      muteDownloadQueue = NULL;
    }
}



char *
muteShareGetLocalVirtualAddress ()
{
  return stringDuplicate (muteShareVirtualAddress);
}



void
muteShareSetSharingPath (char *inPath)
{
  // encode to protect spaces and special characters
  char *safePath = URLUtils::hexEncode (inPath);
  SettingsManager::setSetting ("sharingPath", safePath);
  delete[]safePath;
}



char *
muteShareGetSharingPath ()
{
  //char *safeSetPath = SettingsManager::getStringSetting ("sharingPath");
  char *safeSetPath = strdup(settings->getSharePathSetting ().toLocal8Bit ().data ());
  if (safeSetPath != NULL)
    {
      char *decodedPath = URLUtils::hexDecode (safeSetPath);
      free(safeSetPath);

      return decodedPath;
    }
  else
    {
      return stringDuplicate ("files");
    }
}



void
muteShareSetMaxSimultaneousDownloads (long inLimit)
{
  SettingsManager::setSetting ("maxSimultaneousDownloads", (int) inLimit);

  muteMaxSimultaneousDownloads = inLimit;
}



unsigned long
muteShareGetMaxSimultaneousDownloads ()
{
  char valueFound;
  unsigned long limit =
    SettingsManager::getIntSetting ("maxSimultaneousDownloads",
				    &valueFound);

  if (valueFound)
    {
      return limit;
    }
  else
    {
      // default to no limit
      return 0;
    }
}



class ShareFileInfoWrapper
{

public:

  ShareFileInfoWrapper (): mInfoReceived (false)
  {

  }


   ~ShareFileInfoWrapper ()
  {
  }


  QMutex mLock;

  char *mVirtualAddress;
  char *mFilePath;

  // true if chunk data has already been received and set here
  // for this file
  // used by FileInfoHandler to avoid overwritting data twice
  char mInfoReceived;

  char mFound;
  int mLength;
  int mChunkCount;
  char *mMimeType;

  QSemaphore *mSemaphore;
};



// handler for FileInfo messages, used by muteGetFileInfo
int
muteShare_internalFileInfoHandler (char *inFromAddress,
				   char *inToAddress,
				   char *inBody, void *inExtraArgument)
{

  // is this a FileInfo message?
  // look at second token
  SimpleVector < char *>*tokens = tokenizeString (inBody);
  int numTokens = tokens->size ();

  if (numTokens >= 6)
    {
      char *typeToken = *(tokens->getElement (1));
      if (strcmp ("FileInfo", typeToken) == 0)
	{
	  // unwrap info object from extra arg
	  ShareFileInfoWrapper *fileInfo =
	    (ShareFileInfoWrapper *) inExtraArgument;


	  fileInfo->mLock.lock ();
	  char infoLocked = true;


	  if (fileInfo->mInfoReceived)
	    {
	      // we already received info for this file
	      // ignore this new message

	      fileInfo->mLock.unlock ();
	      infoLocked = false;

	      // no utility in any case
	      return 0;
	    }


	  if (strcmp (inFromAddress, fileInfo->mVirtualAddress) != 0)
	    {
	      // not the address we're looking for

	      fileInfo->mLock.unlock ();
	      infoLocked = false;

	      // no utility in any case
	      return 0;
	    }



	  char *encodedPath = *(tokens->getElement (3));
	  char *filePath = URLUtils::hexDecode (encodedPath);

	  // is this info about our file?
	  if (strcmp (filePath, fileInfo->mFilePath) == 0)
	    {

	      char *statusString = *(tokens->getElement (5));

	      // was file found?
	      if (strcmp (statusString, "Found") == 0)
		{

		  if (numTokens == 12)
		    {
		      fileInfo->mFound = true;

		      // flag so we won't overwrite it a second time
		      fileInfo->mInfoReceived = true;

		      char *lengthString = *(tokens->getElement (7));
		      // default to 0
		      int length = 0;
		      sscanf (lengthString, "%d", &length);

		      fileInfo->mLength = length;

		      char *chunkCountString = *(tokens->getElement (9));
		      // default to 0
		      int chunkCount = 0;
		      sscanf (chunkCountString, "%d", &chunkCount);

		      fileInfo->mChunkCount = chunkCount;

		      fileInfo->mMimeType =
			stringDuplicate (*(tokens->getElement (11)));

		      fileInfo->mLock.unlock ();
		      infoLocked = false;

		      fileInfo->mSemaphore->release ();
		    }
		  // else info formatted incorrectly, ignore

		}
	      else
		{
		  // not found, but we got the info
		  fileInfo->mFound = false;

		  fileInfo->mLock.unlock ();
		  infoLocked = false;

		  fileInfo->mSemaphore->release ();
		}
	    }
	  delete[]filePath;
	  if (infoLocked)
	    {
	      fileInfo->mLock.unlock ();
	      infoLocked = false;
	    }
	}
    }

  for (int i = 0; i < numTokens; i++)
    {
      delete[] * (tokens->getElement (i));
    }

  delete tokens;



  // no utility generated
  return 0;
}



/**
 * Gets whether a download is interrupted and clears the interrupted flag.
 *
 * @param inDownloadID the ID of the download to check.
 */
char
muteShare_internalIsDownloadInterrupted (int inDownloadID)
{
  muteDownloadDataLock.lock ();
  char found = false;
  char returnValue = false;
  for (int i = 0; i < muteDownloadIDs->size () && !found; i++)
    {
      if (inDownloadID == *(muteDownloadIDs->getElement (i)))
	{

	  char *flag = muteDownloadInterruptedFlags->getElement (i);
	  returnValue = *flag;

	  // reset flag
	  *flag = false;
	}
    }

  muteDownloadDataLock.unlock ();

  return returnValue;
}



/**
 * Gets the semaphore associated with a download.
 *
 * @param inDownloadID the ID of the download.
 *
 * @return the semaphore, or NULL if the download is not found.
 *   Should not be destroyed by caller.
 */
QSemaphore *
muteShare_internalGetDownloadSemaphore (int inDownloadID)
{

  muteDownloadDataLock.lock ();
  char found = false;
  QSemaphore *returnValue = NULL;
  for (int i = 0; i < muteDownloadIDs->size () && !found; i++)
    {
      if (inDownloadID == *(muteDownloadIDs->getElement (i)))
	{

	  returnValue = *(muteDownloadSemaphores->getElement (i));
	  found = true;
	}
    }

  muteDownloadDataLock.unlock ();

  return returnValue;
}



/**
 * Gets the source queue associated with a download.
 *
 * Since this call returns heap-allocated data, it is NOT thread-safe.
 * muteDownloadDataLock must be locked by caller to ensure safety.
 *
 * @param inDownloadID the ID of the download.
 *
 * @return the source queue, or NULL if the download is not found.
 *   Should not be destroyed by caller.
 */
SourceQueue *
muteShare_internalGetDownloadSourceQueue (int inDownloadID)
{
  char found = false;
  SourceQueue *returnValue = NULL;
  for (int i = 0; i < muteDownloadIDs->size () && !found; i++)
    {
      if (inDownloadID == *(muteDownloadIDs->getElement (i)))
	{

	  returnValue = *(muteDownloadSourceQueues->getElement (i));
	  found = true;
	}
    }

  return returnValue;
}



/**
 * FIXME FIXME : why to hibernate in such case ???
 * donkey don't make so !!!
 *
 * Puts a download into hibernation mode until more sources are found.
 *
 * @param inDownloadID the ID of the hibernating download.
 * @param inFileChunkHandler the handler to call each time we retry.
 *   This handler follows the same specification as the handler
 *   that is passed into muteShareGetFile().
 * @param inExtraHandlerArgument the extra argument to pass into the handler.
 *
 * Returns only if more sources are found or if download was canceled.
 *
 * @return true if sources found, or false if canceled.
 */
char
muteShare_internalHibernateForMoreSources (int inDownloadID,
					   char (*inFileChunkHandler)
					   (unsigned char *, int, int,
					    void *),
					   void *inExtraHandlerArgument)
{

  SourceQueue *sources =
    muteShare_internalGetDownloadSourceQueue (inDownloadID);

  QSemaphore *downloadSemaphore =
    muteShare_internalGetDownloadSemaphore (inDownloadID);

  // flag queue that we are hibernating
  sources->setHibernating (true);

  // ping handler to give it a chance to notice that we are
  // hibernating
  // watch for canceled signal
  char canceled = !inFileChunkHandler (NULL, -1, -1,
				       inExtraHandlerArgument);



  // 20 minutes
  unsigned long hibernationTime = 1200000;
  // 10 sec for testing
  //unsigned long hibernationTime = 10000;

  // 60 seconds
  unsigned long searchTime = 60000;
  // 10 for testing
  //unsigned long searchTime = 10000;

  int sourceCount = sources->getActiveSourceCount ();

  while (!canceled && sourceCount == 0)
    {
      sources->searchForMoreSources (searchTime);


      // wake up throughout our search to see if results have come in
      unsigned long netSleepTime = 0;

      while (!canceled && sourceCount == 0 && netSleepTime < searchTime)
	{

	  canceled = muteShare_internalIsDownloadInterrupted (inDownloadID);

	  sourceCount = sources->getActiveSourceCount ();

	  if (!canceled && sourceCount == 0)
	    {
	      // sleep on semph so it can be interrupted by
	      // cancel

	      // 10 second sleep chunks during our search
	      long sleepTime = 10000;
	      downloadSemaphore->tryAcquire (1, sleepTime);

	      netSleepTime += sleepTime;
	    }
	}



      if (!canceled)
	{
	  // check again
	  canceled = muteShare_internalIsDownloadInterrupted (inDownloadID);
	}

      sourceCount = sources->getActiveSourceCount ();

      if (!canceled && sourceCount == 0)
	{
	  // hibernate

	  // sleep on semph so it can be interrupted by cancel
	  downloadSemaphore->tryAcquire (1, hibernationTime);
	}
    }

  // if we got to here, we either have new sources, or we
  // have been canceled.

  // flag queue that we are done hibernating
  sources->setHibernating (false);

  return !canceled;
}



int MUTE_SHARE_FILE_INFO_TIMEOUT = 0;
int MUTE_SHARE_FILE_INFO_FOUND = 1;
int MUTE_SHARE_FILE_INFO_NOT_FOUND = 2;
int MUTE_SHARE_FILE_INFO_CANCELED = 3;

/**
 * Gets info about a shared file.
 *
 * @param inQueue the queue to use for download sources.
 * @param outLength pointer to where the length of the file should be
 *   returned.
 * @param outChunkCount pointer to where the number of chunks in the file
 *   should be returned.
 * @param outMimeType pointer to where the mime type of the file should
 *   be returned.  Will be set to NULL if fetching info fails.
 *   Returned string must be destroyed by caller.
 * @param inTimeoutInMilliseconds the starting timeout in milliseconds.
 * @param inDownloadID the ID of the download in our list.
 * @param inFileChunkHandler the handler to call each time we retry.
 *   This handler follows the same specification as the handler
 *   that is passed into muteShareGetFile().
 * @param inExtraHandlerArgument the extra argument to pass into the handler.
 *
 * @return one of MUTE_SHARE_FILE_INFO_TIMEOUT, MUTE_SHARE_FILE_INFO_FOUND,
 *   MUTE_SHARE_FILE_INFO_NOT_FOUND, or MUTE_SHARE_FILE_INFO_CANCELED.
 */
int
muteShare_internalGetFileInfo (SourceQueue * inQueue,
			       int *outLength,
			       int *outChunkCount,
			       char **outMimeType,
			       int inTimeoutInMilliseconds,
			       int inDownloadID,
			       char (*inFileChunkHandler) (unsigned char *,
							   int, int, void *),
			       void *inExtraHandlerArgument)
{

  RandomSource *randSource = new StdRandomSource ();


  char *currentSourceAddress;
  char *currentSourceFilePath;




  char canceled = false;


  char settingFound;
  int maxNumRetries =
    SettingsManager::getIntSetting ("downloadFileInfoRetries",
				    &settingFound);
  if (!settingFound)
    {
      maxNumRetries = 5;
    }

  float freshRouteProbability =
    SettingsManager::getFloatSetting ("downloadRetryFreshRouteProbability",
				      &settingFound);
  if (!settingFound)
    {
      freshRouteProbability = 0.25;
    }

  // double timeout for a retry 
  int retryTimeoutMilliseconds = 2 * inTimeoutInMilliseconds;


  ShareFileInfoWrapper *fileInfo = new ShareFileInfoWrapper ();

  // get next source
  inQueue->getNextSource (&currentSourceAddress, &currentSourceFilePath);
  fileInfo->mVirtualAddress = currentSourceAddress;
  fileInfo->mFilePath = currentSourceFilePath;

  fileInfo->mSemaphore =
    muteShare_internalGetDownloadSemaphore (inDownloadID);
  fileInfo->mMimeType = NULL;

  // register a handler
  int handlerID = muteAddMessageHandler (muteShare_internalFileInfoHandler,
					 (void *) fileInfo);

  // send out a request message
  char *encodedFilePath = URLUtils::hexEncode (currentSourceFilePath);

  char *message = autoSprintf ("MessageType: FileInfoRequest\n"
			       "FilePath: %s",
			       encodedFilePath);
  delete[]encodedFilePath;

  QTime startTime;
  startTime.start ();

  muteSendMessage (muteShareVirtualAddress, currentSourceAddress, message);

  delete[]message;



  int returnValue;

  // wait for our handler to get the response
  int responseReceived =
    fileInfo->mSemaphore->tryAcquire (1, inTimeoutInMilliseconds);

  unsigned long responseTime = startTime.elapsed ();

  if (muteShare_internalIsDownloadInterrupted (inDownloadID))
    {
      // interrupted... response not really received.
      responseReceived = 0;
    }
  else if (responseReceived)
    {
      // report response time to source queue
      inQueue->reportBestSourceChunkTime (responseTime);
    }
  else
    {
      // report failure
      inQueue->reportBestSourceChunkTime (-1);
    }



  int numRetries = 0;

  while (responseReceived != 1 && numRetries < maxNumRetries && !canceled)
    {

      // no response, so retry

      // first ask chunk handler if we should keep going
      // pass NULL to it to indicate that we timed out on the current
      // info request
      char keepGoing = inFileChunkHandler (NULL, -1, -1,
					   inExtraHandlerArgument);


      if (!keepGoing)
	{
	  canceled = true;
	}
      else
	{
	  // another retry

	  // lock the file chunk to avoid access to bad data by handler
	  // as we switch sources
	  fileInfo->mLock.lock ();

	  // get next source
	  delete[]currentSourceAddress;
	  currentSourceAddress=NULL;
	  delete[]currentSourceFilePath;
	  currentSourceFilePath=NULL;
	  inQueue->getNextSource (&currentSourceAddress,
				  &currentSourceFilePath);
	  fileInfo->mVirtualAddress = currentSourceAddress;
	  fileInfo->mFilePath = currentSourceFilePath;

	  fileInfo->mLock.unlock ();


	  // should we use a freshroute?
	  // pick at random,
	  // but always use a freshroute on the last retry
	  char useFreshRoute = false;
	  if (randSource->getRandomFloat () <= freshRouteProbability)
	    {
	      useFreshRoute = true;
	    }

	  if (numRetries == maxNumRetries - 1)
	    {
	      // last retry
	      // go down fighting with a FRESH_ROUTE
	      useFreshRoute = true;
	    }

	  char *flags;
	  char *extraLogMessage;

	  if (useFreshRoute)
	    {
	      flags = "FRESH_ROUTE";
	      extraLogMessage = "and trying a FRESH_ROUTE";
	    }
	  else
	    {
	      flags = NULL;
	      extraLogMessage = "";
	    }



	  // retry with a fresh route flag
	  char *logMessage =
	    autoSprintf
	    ("Timed out, so sending retry %d with timeout of %d ms, "
	     "%s to get info for file %s : %s",
	     numRetries, retryTimeoutMilliseconds, extraLogMessage,
	     currentSourceAddress, currentSourceFilePath);
	  AppLog::info ("fileShare", logMessage);
	  delete[]logMessage;


	  // reformat message in case source (and thus file path) has changed
	  encodedFilePath = URLUtils::hexEncode (currentSourceFilePath);

	  message = autoSprintf ("MessageType: FileInfoRequest\n"
				 "FilePath: %s", encodedFilePath);
	  delete[]encodedFilePath;


	  startTime.start ();

	  muteSendMessage (muteShareVirtualAddress,
			   currentSourceAddress, message, flags);

	  delete[]message;


	  // wait a second time for our handler to get the response
	  responseReceived =
	    fileInfo->mSemaphore->tryAcquire (1, retryTimeoutMilliseconds);

	  responseTime = startTime.elapsed ();

	  if (muteShare_internalIsDownloadInterrupted (inDownloadID))
	    {
	      // interrupted... response not really received.
	      responseReceived = 0;
	    }
	  else if (responseReceived)
	    {
	      // report response time to source queue
	      inQueue->reportBestSourceChunkTime (responseTime);
	    }
	  else
	    {
	      // report failure
	      inQueue->reportBestSourceChunkTime (-1);
	    }


	  numRetries++;


	  if (!responseReceived && !canceled && numRetries == maxNumRetries)
	    {

	      // reset retry count to start over after we find more
	      // sources
	      numRetries = 0;

	      // result of hibernation either found sources or canceled
	      canceled =
		!muteShare_internalHibernateForMoreSources (inDownloadID,
							    inFileChunkHandler,
							    inExtraHandlerArgument);
	    }
	}
    }


  muteRemoveMessageHandler (handlerID);


  if (canceled)
    {
      char *logMessage = autoSprintf ("Canceled while getting info for "
				      "file %s : %s",
				      currentSourceAddress,
				      currentSourceFilePath);
      AppLog::info ("fileShare", logMessage);
      delete[]logMessage;

      *outMimeType = NULL;
      returnValue = MUTE_SHARE_FILE_INFO_CANCELED;

      if (fileInfo->mMimeType != NULL)
	{
	  // maybe handler was called between our cancel and
	  // our call to muteRemoveMessageHandler

	  // ignore it anyway, but clean up
	  delete[]fileInfo->mMimeType;
	  fileInfo->mMimeType = NULL;
	}
    }
  else if (responseReceived == 1)
    {

      if (fileInfo->mFound)
	{
	  *outLength = fileInfo->mLength;
	  *outChunkCount = fileInfo->mChunkCount;
	  *outMimeType = fileInfo->mMimeType;

	  returnValue = MUTE_SHARE_FILE_INFO_FOUND;
	}
      else
	{
	  *outMimeType = NULL;
	  returnValue = MUTE_SHARE_FILE_INFO_NOT_FOUND;
	}
    }
  else
    {
      // failed

      char *logMessage =
	autoSprintf
	("Giving up:  timed out (after %d retries, each with a timeout "
	 "of %d ms) getting info for " "file %s : %s",
	 maxNumRetries,
	 retryTimeoutMilliseconds,
	 currentSourceAddress, currentSourceFilePath);
      AppLog::warning ("fileShare", logMessage);
      delete[]logMessage;

      *outMimeType = NULL;
      returnValue = MUTE_SHARE_FILE_INFO_TIMEOUT;

      if (fileInfo->mMimeType != NULL)
	{
	  // maybe handler was called between our timeout and
	  // our call to muteRemoveMessageHandler

	  // ignore it anyway, but clean up
	  delete[]fileInfo->mMimeType;
	  fileInfo->mMimeType = NULL;
	}
    }


  delete[]currentSourceAddress;
  delete[]currentSourceFilePath;

  delete fileInfo;

  delete randSource;

  return returnValue;
}



class ShareSearchWrapper
{

public:
  char *mSearchID;

  QMutex mLock;
  char mCanceled;
  char *mFromAddress;
  int mNumResults;
  char **mFilePaths;
  unsigned long *mLengthsInBytes;
  char **mHashes;

  QSemaphore *mResultsReadySemaphore;

  QSemaphore *mResultsConsumedSemaphore;
};



// handler for SearchResults messages, used by muteShareSearch
int
muteShare_internalSearchResultsHandler (char *inFromAddress,
					char *inToAddress,
					char *inBody, void *inExtraArgument)
{
  // unwrap info object from extra arg
  ShareSearchWrapper *search = (ShareSearchWrapper *) inExtraArgument;

  // is this a SearchResults message?
  // look at second token
  SimpleVector < char *>*tokens = tokenizeString (inBody);
  int numTokens = tokens->size ();

  if (numTokens >= 6)
    {
      char *typeToken = *(tokens->getElement (1));
      if (strcmp ("SearchResults", typeToken) == 0)
	{


	  char *searchID = *(tokens->getElement (3));

	  // are these results for our search?
	  if (strcmp (searchID, search->mSearchID) == 0)
	    {

	      char *resultCountString = *(tokens->getElement (5));
	      // default to 0
	      int resultCount = 0;
	      sscanf (resultCountString, "%d", &resultCount);

	      int firstResultToken = 7;
	      // 3 tokens per result
	      int lastResultToken = firstResultToken + 3 * resultCount;

	      if (numTokens >= lastResultToken)
		{
		  if (muteSharePrintSearchSyncTrace)
		    {
		      printf ("  Result handler locking search mutex\n");
		    }
		  search->mLock.lock ();

		  if (!search->mCanceled)
		    {

		      search->mFromAddress = stringDuplicate (inFromAddress);
		      search->mNumResults = resultCount;
		      search->mFilePaths = new char *[resultCount];
		      search->mLengthsInBytes =
			new unsigned long[resultCount];
		      search->mHashes = new char *[resultCount];

		      int resultIndex = 0;
		      for (int i = firstResultToken;
			   i < lastResultToken - 2; i += 3)
			{

			  char *encodedFileName = *(tokens->getElement (i));
			  char *lengthString = *(tokens->getElement (i + 1));
			  char *hashString = *(tokens->getElement (i + 2));

			  search->mFilePaths[resultIndex] =
			    URLUtils::hexDecode (encodedFileName);

			  // default to 0 if scan fails
			  unsigned long length = 0;
			  sscanf (lengthString, "%lu", &length);
			  search->mLengthsInBytes[resultIndex] = length;

			  search->mHashes[resultIndex] =
			    stringDuplicate (hashString);


			  resultIndex++;
			}

		      if (muteSharePrintSearchSyncTrace)
			{
			  printf
			    ("  Result handler unlocking search mutex\n");
			}
		      search->mLock.unlock ();

		      // signal that results are ready
		      if (muteSharePrintSearchSyncTrace)
			{
			  printf ("  Result handler signaling that "
				  "results are ready\n");
			}
		      search->mResultsReadySemaphore->release ();

		      // wait for results to be consumed
		      if (muteSharePrintSearchSyncTrace)
			{
			  printf ("  Result handler waiting for "
				  "results to be consumed\n");
			}
		      search->mResultsConsumedSemaphore->tryAcquire ();
		      if (muteSharePrintSearchSyncTrace)
			{
			  printf ("  Result handler done waiting for "
				  "results to be consumed\n");
			}

		    }
		  else
		    {
		      // canceled, don't return results
		      search->mLock.unlock ();
		    }
		}
	    }
	}
    }

  for (int i = 0; i < numTokens; i++)
    {
      delete[] * (tokens->getElement (i));
    }

  delete tokens;

  // no utility generated
  return 0;
}






void
muteShareSearch (char *inSearchString,
		 char (*inResultHandler) (char *, char *, unsigned long,
					  char *, void *),
		 void *inExtraHandlerArgument, int inTimeoutInMilliseconds,
		 char *inAddressToSearch)
{

  char *logMessage = autoSprintf ("Searching for %s", inSearchString);
  AppLog::info ("fileShare", logMessage);
  delete[]logMessage;


  char *searchID = muteGetUniqueName ();

  // register a handler for the chunk
  ShareSearchWrapper *search = new ShareSearchWrapper ();
  search->mSearchID = searchID;
  search->mCanceled = false;
  search->mFromAddress = NULL;
  search->mNumResults = 0;
  search->mFilePaths = NULL;
  search->mLengthsInBytes = NULL;
  search->mHashes = NULL;

  search->mResultsReadySemaphore = new QSemaphore ();
  search->mResultsConsumedSemaphore = new QSemaphore ();

  int handlerID =
    muteAddMessageHandler (muteShare_internalSearchResultsHandler,
			   (void *) search);


  // send out our search
  char *encodedSearchString = URLUtils::hexEncode (inSearchString);

  char *message = autoSprintf ("MessageType: SearchRequest\n"
			       "SearchID: %s\n" "SearchString: %s",
			       searchID,
			       encodedSearchString);

  char *forwardFlag;

  if (strcmp (inAddressToSearch, "ALL") == 0)
    {
      // sending to ALL
      // use the FORWARD flag to preserve our anonymity
      char *hashSeed = muteGetForwardHashSeed ();
      forwardFlag = autoSprintf ("FORWARD_%s", hashSeed);
      delete[]hashSeed;
    }
  else
    {
      // sending to one, no FORWARD flag needed
      forwardFlag = stringDuplicate ("");
    }

  muteSendMessage (muteShareVirtualAddress,
		   inAddressToSearch, message, forwardFlag);
  delete[]forwardFlag;
  delete[]message;

  delete[]encodedSearchString;



  char canceled = false;

  while (!canceled)
    {

      // wait for results
      if (muteSharePrintSearchSyncTrace)
	{
	  printf ("Search waiting for " "results to be ready\n");
	}
      int resultsReceived =
	search->mResultsReadySemaphore->tryAcquire (1,
						    inTimeoutInMilliseconds);

      if (muteSharePrintSearchSyncTrace)
	{
	  printf ("Search done waiting for " "results to be ready\n");
	}


      if (resultsReceived == 1)
	{
	  // pass the results to the caller's handler

	  if (muteSharePrintSearchSyncTrace)
	    {
	      printf ("Search locking search mutex\n");
	    }
	  search->mLock.lock ();

	  int i;
	  for (i = 0; i < search->mNumResults && !canceled; i++)
	    {
	      char keepGoing = inResultHandler (search->mFromAddress,
						search->mFilePaths[i],
						search->mLengthsInBytes[i],
						search->mHashes[i],
						inExtraHandlerArgument);

	      if (!keepGoing)
		{
		  canceled = true;
		}
	    }

	  // delete results
	  delete[]search->mFromAddress;
	  search->mFromAddress = NULL;
	  for (i = 0; i < search->mNumResults; i++)
	    {
	      delete[]search->mFilePaths[i];
	      delete[]search->mHashes[i];
	    }
	  delete[]search->mFilePaths;
	  delete[]search->mLengthsInBytes;
	  delete[]search->mHashes;

	  search->mFilePaths = NULL;
	  search->mLengthsInBytes = NULL;
	  search->mHashes = NULL;

	  search->mNumResults = 0;

	  search->mCanceled = canceled;

	  if (muteSharePrintSearchSyncTrace)
	    {
	      printf ("Search unlocking search mutex\n");
	    }
	  search->mLock.unlock ();

	  if (!canceled)
	    {
	      // signal that results have been consumed
	      if (muteSharePrintSearchSyncTrace)
		{
		  printf
		    ("Search signaling that results have been consumed\n");
		}
	      search->mResultsConsumedSemaphore->release ();
	    }
	  // else keep handler blocked until we can remove it below
	}
      else
	{
	  // timed out waiting

	  // tell handler that we timed out
	  char keepGoing = inResultHandler (NULL,
					    NULL,
					    0,
					    NULL,
					    inExtraHandlerArgument);

	  if (!keepGoing)
	    {
	      canceled = true;

	      search->mLock.lock ();
	      search->mCanceled = true;
	      search->mLock.unlock ();
	    }

	}
    }


  // handler has canceled us


  // delete any extra results
  if (muteSharePrintSearchSyncTrace)
    {
      printf ("Search locking search mutex\n");
    }
  search->mLock.lock ();
  if (search->mFromAddress != NULL)
    {
      delete[]search->mFromAddress;
      search->mFromAddress = NULL;

      for (int i = 0; i < search->mNumResults; i++)
	{
	  delete[]search->mFilePaths[i];
	  delete[]search->mHashes[i];
	}
      delete[]search->mFromAddress;
      delete[]search->mLengthsInBytes;
      delete[]search->mHashes;

      search->mFromAddress = NULL;
      search->mLengthsInBytes = NULL;
      search->mHashes = NULL;
    }
  search->mCanceled = true;
  if (muteSharePrintSearchSyncTrace)
    {
      printf ("Search unlocking search mutex\n");
    }
  search->mLock.unlock ();

  if (muteSharePrintSearchSyncTrace)
    {
      printf ("Search signaling that results have been consumed\n");
    }
  search->mResultsConsumedSemaphore->release ();


  // remove the result handler
  muteRemoveMessageHandler (handlerID);

  delete search->mResultsReadySemaphore;
  delete search->mResultsConsumedSemaphore;

  delete[]search->mSearchID;

  delete search;
}



class ShareFileChunkWrapper
{

public:


  ShareFileChunkWrapper ()
  {
  }


   ~ShareFileChunkWrapper ()
  {

  }

  char *mVirtualAddress;
  char *mFilePath;

  int mChunkNumber;

  // true if chunk data has already been received and set here
  // for mChunkNumber
  // used by FileChunkHandler to avoid overwritting data twice
  char mDataReceived;

  int mLengthInBytes;
  unsigned char *mChunkData;

  QMutex mLock;
  QSemaphore *mSemaphore;
};



// handler for FileChunk messages, used by muteShareGetFile
int
muteShare_internalFileChunkHandler (char *inFromAddress,
				    char *inToAddress,
				    char *inBody, void *inExtraArgument)
{

  // is this a FileChunk message?
  // look at second token
  SimpleVector < char *>*tokens = tokenizeString (inBody);
  int numTokens = tokens->size ();

  if (numTokens >= 10)
    {
      char *typeToken = *(tokens->getElement (1));
      if (strcmp ("FileChunk", typeToken) == 0)
	{
//printf("Chunk Received : %s\n",*(tokens->getElement (3)));
	  // unwrap info object from extra arg
	  ShareFileChunkWrapper *fileChunk =
	    (ShareFileChunkWrapper *) inExtraArgument;


	  // make sure data in file chunk is not changed by another thread now 
	  fileChunk->mLock.lock ();
	  char stillLocked = true;

	  if (fileChunk->mDataReceived)
	    {
	      // we already received data for our chunk
	      // ignore this new message

	      // no utility in any case
	      fileChunk->mLock.unlock ();
	      stillLocked = false;
	      goto no_utility;
	    }

	  if (strcmp (inFromAddress, fileChunk->mVirtualAddress) != 0)
	    {
	      // not the address we're looking for

	      // no utility in any case
	      fileChunk->mLock.unlock ();
	      stillLocked = false;
	      goto no_utility;
	    }


	  char *encodedPath = *(tokens->getElement (3));
	  char *filePath = URLUtils::hexDecode (encodedPath);

	  // is this info about our file?
	  if (strcmp (filePath, fileChunk->mFilePath) == 0)
	    {

	      char *chunkNumberString = *(tokens->getElement (5));
	      // default to 0
	      int chunkNumber = 0;
	      sscanf (chunkNumberString, "%d", &chunkNumber);

	      // is this the chunk we're looking for?
	      if (fileChunk->mChunkNumber == chunkNumber)
		{

		  char *lengthString = *(tokens->getElement (7));
		  // default to 0
		  int length = 0;
		  sscanf (lengthString, "%d", &length);

		  fileChunk->mLengthInBytes = length;

		  char *encodedData = *(tokens->getElement (9));
		  int dataLength;

		  unsigned char *decodedData =
		    base64Decode (encodedData, &dataLength);

		  if (decodedData != NULL)
		    {

		      if (dataLength == length)
			{

			  fileChunk->mChunkData = decodedData;

			  // we have now received this chunk, so we
			  // should never overwrite it again if another
			  // message comes in
			  fileChunk->mDataReceived = true;

			  // unlock before signal
			  fileChunk->mLock.unlock ();
			  stillLocked = false;

			  fileChunk->mSemaphore->release ();
			}
		      else
			{
			  AppLog::error ("fileShare -- chunk handler",
					 "Data length incorrect");

			  delete[]decodedData;
			}
		    }
		  else
		    {
		      AppLog::error ("fileShare -- chunk handler",
				     "Failed to decode data (base64 format bad?)");
		    }

		}
	    }
	  delete[]filePath;
	  if (stillLocked)
	    {
	      // still need to unlock
	      fileChunk->mLock.unlock ();
	      stillLocked = false;
	    }
	}
    }

no_utility:
  for (int i = 0; i < numTokens; i++)
    {
      delete[] * (tokens->getElement (i));
    }

  delete tokens;



  // no utility generated
  return 0;
}



/**
 * Removes info from our list about a download.
 *
 * Does not delete the semaphore associated with the download (simply
 * removes it from the list.
 * Does not delete the source queue.
 *
 * @param inDownloadID the ID to remove.
 */
void
muteShare_internalRemoveDownloadInfo (int inDownloadID)
{
  // remove download info from our list
  muteDownloadDataLock.lock ();
  char found = false;
  for (int i = 0; i < muteDownloadIDs->size () && !found; i++)
    {
      if (inDownloadID == *(muteDownloadIDs->getElement (i)))
	{

	  muteDownloadIDs->deleteElement (i);
	  muteDownloadSemaphores->deleteElement (i);
	  muteDownloadInterruptedFlags->deleteElement (i);
	  muteDownloadSourceQueues->deleteElement (i);
	  found = true;
	}
    }

  muteDownloadDataLock.unlock ();
}



/**
 * Thread that searches for more sources and adds them to a SourceQueue.
 *
 * @author Jason Rohrer.
 */
class SourceSearchThread:public FinishedSignalThread
{

public:

	/**
         * Constructs and starts a search thread.
         *
         * @param inQueue the queue to add results to.
         *   Destroyed by caller.
         * @param inSearchTimeoutInMilliseconds how long the search waits
         *   for responses before timing out.
         */
  SourceSearchThread (SourceQueue * inQueue,
		      unsigned long inSearchTimeoutInMilliseconds);


	/**
         * Joins and destroys a search thread.
         */
   ~SourceSearchThread ();



	/**
         * Called by result handler to tell us that a new result has arrived.
         *
         * @return true if search should continue, or false if it should stop.
         */
  char gotResult ();


  // implements Thread interface
  void run ();

protected:
    SourceQueue * mQueue;

  unsigned long mSearchTimeoutInMilliseconds;

  QTime mSearchStart;

  QMutex mLock;
  char mContinue;
};



SourceQueue::SourceQueue (char *inFileHash,
			  char *inFirstVirtualAddress, char *inFirstFilePath):
mHibernating (false),
mLastThread (NULL),
mFileHash (stringDuplicate (inFileHash)),
mSourceVector (new SimpleVector < FileSource * >()),
mLongestSourceList (1)
{

  FileSource *firstSource =
    new FileSource (stringDuplicate (inFirstVirtualAddress),
		    stringDuplicate (inFirstFilePath));
  mSourceVector->push_back (firstSource);

  // init time
  mLastSearchStart.start ();
}



SourceQueue::~SourceQueue ()
{

  if (mLastThread != NULL)
    {
      // join and destroy the thread
      delete mLastThread;
      mLastThread = NULL;
    }

  delete[]mFileHash;


  int numSources = mSourceVector->size ();

  for (int i = 0; i < numSources; i++)
    {
      delete *(mSourceVector->getElement (i));
    }
  delete mSourceVector;

}




void
SourceQueue::reportBestSourceChunkTime (long inMilliseconds)
{
  mLock.lock ();

  FileSource *bestSource = *(mSourceVector->getElement (0));

  // set new time
  bestSource->mChunkTimeInMilliseconds = inMilliseconds;

  mLock.unlock ();


  int activeCount = getActiveSourceCount ();

  unsigned long timeSinceLastSearch = mLastSearchStart.elapsed ();

  // if
  // our list is empty, or
  // we're down to half of our largest list, or
  // we haven't searched in more than 10 minutes

  if (activeCount == 0 ||
      activeCount <= mLongestSourceList / 2 ||
      timeSinceLastSearch > 10 * 60000)
    {

      // add more
      // search for 30 seconds

      // note start time
      mLastSearchStart.start ();

      searchForMoreSources (30000);
    }


  // update the current list
  // always do this, whether we started a new search or not,
  // because the search will take a while to update our source list
  findNewBest ();
}



void
SourceQueue::getNextSource (char **outVirtualAddress, char **outFilePath)
{
  mLock.lock ();

  FileSource *bestSource = *(mSourceVector->getElement (0));

  *outVirtualAddress = stringDuplicate (bestSource->mVirtualAddress);
  *outFilePath = stringDuplicate (bestSource->mFilePath);

  mLock.unlock ();
}



int
SourceQueue::getActiveSourceCount ()
{
  mLock.lock ();

  // count how many we have that are not marked with failure

  int numSources = mSourceVector->size ();

  int notFailedCount = 0;
  int index = 0;
  for (index = 0; index < numSources; index++)
    {
      FileSource *otherSource = *(mSourceVector->getElement (index));

      if (otherSource->mChunkTimeInMilliseconds != -1)
	{
	  notFailedCount++;
	}
    }

  mLock.unlock ();

  return notFailedCount;
}



/**
 * The extra argument expected by searchForMoreSourcesResultHandler.
 */
class SearchForMoreSourcesResultHandler_ExtraArgument
{
public:
  SourceSearchThread * mThread;
  SourceQueue *mSourceQueue;
};



/**
 * The handler passed into muteShareSearch by searchForMoreSources.
 */
char
searchForMoreSourcesResultHandler (char *inFileHostVirtualAddress,
				   char *inFilePath,
				   unsigned long inFileLengthInBytes,
				   char *inFileSHA1Hash,
				   void *inExtraArgument)
{

  SearchForMoreSourcesResultHandler_ExtraArgument *argument =
    (SearchForMoreSourcesResultHandler_ExtraArgument *) inExtraArgument;

  // extra argument contains our source queue and our thread
  SourceQueue *sources = argument->mSourceQueue;

  SourceSearchThread *thread = argument->mThread;


  if (inFileHostVirtualAddress != NULL)
    {
      // a result

      // add to queue
      FileSource *newSource =
	new FileSource (stringDuplicate (inFileHostVirtualAddress),
			stringDuplicate (inFilePath));

      sources->addSource (newSource);

      delete newSource;
    }
  // else timed out


  char shouldContinue = thread->gotResult ();

  return shouldContinue;
}



SourceSearchThread::SourceSearchThread (SourceQueue * inQueue, unsigned long inSearchTimeoutInMilliseconds):mQueue (inQueue),
mSearchTimeoutInMilliseconds (inSearchTimeoutInMilliseconds),
mContinue (true)
{

  mSearchStart.start ();

  start ();
}



SourceSearchThread::~SourceSearchThread ()
{
  mLock.lock ();
  mContinue = false;
  mLock.unlock ();

  this->terminate ();
  this->wait (20000);

}



char
SourceSearchThread::gotResult ()
{

  // check if we have waited long enough for results
  unsigned long searchTimeSoFar = mSearchStart.elapsed ();

  mLock.lock ();

  if (searchTimeSoFar > mSearchTimeoutInMilliseconds)
    {
      mContinue = false;
    }

  char shouldContinue = mContinue;
  mLock.unlock ();

  return shouldContinue;
}



void
SourceSearchThread::run ()
{
  mQueue->searchForMoreSources_blocking (this);

  setFinished ();
}



char
SourceQueue::
searchForMoreSources (unsigned long inSearchTimeoutInMilliseconds)
{

  mLock.lock ();

  if (mLastThread != NULL)
    {

      if (!mLastThread->isFinished ())
	{

	  mLock.unlock ();

	  // another search is running already
	  return false;
	}
      else
	{

	  // there is another thread that has been started
	  // but it is done

	  // thread already finished, so this shouldn't block
	  delete mLastThread;
	  mLastThread = NULL;
	}
    }

  // reset count of longest source list
  // thus, we are counting active sources found right now
  mLongestSourceList = 0;

  // start a new thread
  mLastThread = new SourceSearchThread (this, inSearchTimeoutInMilliseconds);

  mLock.unlock ();

  return true;
}



void
SourceQueue::searchForMoreSources_blocking (SourceSearchThread * inThread)
{

  char *searchString = autoSprintf ("hash_%s", mFileHash);

  SearchForMoreSourcesResultHandler_ExtraArgument *argument =
    new SearchForMoreSourcesResultHandler_ExtraArgument ();

  argument->mThread = inThread;
  argument->mSourceQueue = this;

  muteShareSearch (searchString, searchForMoreSourcesResultHandler,
		   // wrapper class as extra argument
		   (void *) argument,
		   // 1 second timeout between result pings
		   1000);

  delete[]searchString;
  delete argument;
}



void
SourceQueue::addSource (FileSource * inSource)
{
  mLock.lock ();


  int numExistingSources = mSourceVector->size ();

  char found = false;
  int foundIndex = -1;



  int index = 0;
  for (index = 0; index < numExistingSources && !found; index++)
    {
      FileSource *otherSource = *(mSourceVector->getElement (index));

      if (strcmp (inSource->mVirtualAddress,
		  otherSource->mVirtualAddress) == 0)
	{
	  found = true;
	  foundIndex = index;
	}
    }

  if (!found)
    {
      // add to vector

      FileSource *copiedSource =
	new FileSource (stringDuplicate (inSource->mVirtualAddress),
			stringDuplicate (inSource->mFilePath));

      mSourceVector->push_back (copiedSource);
    }
  else
    {
      // got result from a source already in our queue
      FileSource *existingSource = *(mSourceVector->getElement (foundIndex));

      if (existingSource->mChunkTimeInMilliseconds == -1)
	{
	  // last report from this source was failure

	  // try this source again, though, since we got results back
	  // again
	  existingSource->mChunkTimeInMilliseconds = 0;
	}
    }

  // add to count of sources found right now
  mLongestSourceList++;

  mLock.unlock ();

  findNewBest ();
}



void
SourceQueue::findNewBest ()
{
  mLock.lock ();

  FileSource *bestSource = *(mSourceVector->getElement (0));

  long oldBestTime = bestSource->mChunkTimeInMilliseconds;

  // check to see if another source is faster

  int numSources = mSourceVector->size ();

  int fastestIndex = -1;
  long fastestTime = oldBestTime;

  for (int i = 0; i < numSources; i++)
    {
      FileSource *otherSource = *(mSourceVector->getElement (i));

      long otherTime = otherSource->mChunkTimeInMilliseconds;

      if (fastestTime != -1)
	{
	  if (otherTime != -1 && otherTime < fastestTime)
	    {

	      fastestIndex = i;
	      fastestTime = otherTime;
	    }
	}
      else
	{
	  if (otherTime != -1)
	    {
	      // any non-failure time is better than failure
	      fastestIndex = i;
	      fastestTime = otherTime;
	    }
	}
    }

  if (fastestIndex != -1)
    {
      // found faster

      // swap
      FileSource *otherSource = *(mSourceVector->getElement (fastestIndex));

      *(mSourceVector->getElement (0)) = otherSource;

      *(mSourceVector->getElement (fastestIndex)) = bestSource;
    }
  else
    {
      // none found that were faster

      if (fastestTime == -1)
	{
	  // all must have been -1 (failed last chunk)

	  int numSources = mSourceVector->size ();
	  if (numSources > 1)
	    {
	      // pick one at random to swap with
	      RandomSource *randSource = new StdRandomSource ();

	      int index = randSource->getRandomBoundedInt (1,
							   numSources - 1);

	      delete randSource;


	      FileSource *otherSource = *(mSourceVector->getElement (index));

	      *(mSourceVector->getElement (0)) = otherSource;

	      *(mSourceVector->getElement (index)) = bestSource;
	    }
	}
    }

  mLock.unlock ();
}



int
muteShareGetFile (char *inVirtualAddress, char *inFilePath,
		  char *inFileHash,
		  char (*inFileChunkHandler) (unsigned char *, int, int,
					      void *),
		  void *inExtraHandlerArgument, int *outDownloadID,
		  int *outFileSizeInBytes, int inTimeoutInMilliseconds,
		  char *inPartialFilePath, int inPartialFileSizeInBytes)
{


  QSemaphore *downloadSemaphore = new QSemaphore ();

  SourceQueue *sources =
    new SourceQueue (inFileHash, inVirtualAddress, inFilePath);


  // add download info to our list
  muteDownloadDataLock.lock ();

  int downloadID = muteNextDownloadID;
  // pass ID to our caller
  *outDownloadID = downloadID;
  muteNextDownloadID++;

  muteDownloadIDs->push_back (downloadID);
  muteDownloadSemaphores->push_back (downloadSemaphore);
  muteDownloadInterruptedFlags->push_back (false);
  muteDownloadSourceQueues->push_back (sources);
  muteDownloadDataLock.unlock ();



  char *logMessage =
    autoSprintf ("Trying to get file %s : %s", inVirtualAddress, inFilePath);
  AppLog::info ("fileShare", logMessage);
  delete[]logMessage;


  // We adjust download chunk timeouts with the following weighted average
  // formula:
  // currentTimeout =
  //     ( A * (2 * lastChunkTime) + B * currentTimeout ) / (A+B);
  // A is the weight given to the most recent chunk round-trip time
  double recentChunkTimeWeight;
  // B is the weight given to our current timeout
  double currentTimeoutWeight;
  // If the round-trip time of recent chunks is consistent, then our
  // timeout will eventually approach (2 *) that average round-trip time.

  char settingFound;
  recentChunkTimeWeight =
    SettingsManager::getFloatSetting ("downloadTimeoutRecentChunkWeight",
				      &settingFound);
  if (!settingFound)
    {
      recentChunkTimeWeight = 1;
    }

  currentTimeoutWeight =
    SettingsManager::getFloatSetting ("downloadTimeoutCurrentTimeoutWeight",
				      &settingFound);

  if (!settingFound)
    {
      currentTimeoutWeight = 2;
    }


  int maxNumRetries = SettingsManager::getIntSetting ("downloadChunkRetries",
						      &settingFound);
  if (!settingFound)
    {
      maxNumRetries = 10;
    }

  float freshRouteProbability =
    SettingsManager::getFloatSetting ("downloadRetryFreshRouteProbability",
				      &settingFound);
  if (!settingFound)
    {
      freshRouteProbability = 0.25;
    }

  int logTimeoutFlag =
    SettingsManager::getIntSetting ("logDownloadTimeoutChanges",
				    &settingFound);

  char logTimeoutChanges;
  FILE *timeoutLogFile;

  if (settingFound && logTimeoutFlag == 1)
    {

      logTimeoutChanges = true;

      // get the pathless file name to use in our timeout log file name
      char *lastSlash = inFilePath;
      char *nextSlash = inFilePath;
      while (nextSlash != NULL)
	{
	  // skip slash
	  if (nextSlash[0] == '/')
	    {
	      lastSlash = &(nextSlash[1]);
	    }

	  nextSlash = strstr (lastSlash, "/");
	}

      char *localFileName = lastSlash;

      char *timeoutLogFileName = autoSprintf ("timeoutLog_%s",
					      localFileName);

      timeoutLogFile = fopen (timeoutLogFileName, "w");

      delete[]timeoutLogFileName;
    }
  else
    {
      // default
      logTimeoutChanges = false;
    }

  unsigned long currentTimeoutMilliseconds = inTimeoutInMilliseconds;

  // the time used by retries
  // (will always be twice currentTimeoutMilliseconds)
  unsigned long retryTimeoutMilliseconds = 2 * currentTimeoutMilliseconds;

  // large value to start, an hour
  // will get replaced by time from first chunk
  unsigned long bestChunkTimeMilliseconds = 3600000;


  // the longest timeout ever used during this transfer
  // we should double this timeout for our final retry 
  unsigned long longestTimeoutMilliseconds = currentTimeoutMilliseconds;

  // flag that we set to try a fresh route message for the next chunk
  // we set this if chunk times are getting longer and longer, which
  // might mean that our route is deteriorating from too many local
  // patch-ups
  char useFreshRouteForNextChunk = false;


  // check if we should be queued
  char queued = false;
  muteDownloadDataLock.lock ();

  if (muteNumActiveDownloads >= muteMaxSimultaneousDownloads)
    {
      queued = true;

      // add ourself to the queue
      muteDownloadQueue->push_back (downloadID);
    }
  else
    {
      // not queued, so increase active count
      muteNumActiveDownloads++;
    }

  muteDownloadDataLock.unlock ();


  if (queued)
    {
      // tell handler that we are queued
      char handlerResponse = inFileChunkHandler (NULL, -2, -1,
						 inExtraHandlerArgument);

      while (queued && handlerResponse && handlerResponse != 2)
	{
	  // not canceled and not force dequeued

	  // wait, but in an interruptable way
	  downloadSemaphore->tryAcquire (1, 5000);

	  // check if there is a slot for us now

	  muteDownloadDataLock.lock ();

	  // there is only possibly a slot for us if we are first in queue
	  if (muteDownloadQueue->getElementIndex ((long) downloadID) == 0)
	    {

	      // we are first
	      if (muteNumActiveDownloads < muteMaxSimultaneousDownloads)
		{
		  // there is a slot

		  queued = false;

		  // we are active now
		  muteNumActiveDownloads++;

		  muteDownloadQueue->deleteElement (0);
		}
	    }
	  muteDownloadDataLock.unlock ();

	  if (queued)
	    {
	      // tell handler that we are still queued
	      handlerResponse = inFileChunkHandler (NULL, -2, -1,
						    inExtraHandlerArgument);
	    }
	}

      if (!handlerResponse)
	{
	  // canceled

	  if (queued)
	    {
	      // still queued

	      // remove ourself from queue
	      muteDownloadDataLock.lock ();
	      muteDownloadQueue->deleteElement (muteDownloadQueue->
						getElementIndex ((long)
								 downloadID));
	      muteDownloadDataLock.unlock ();
	    }

	  // clean up download info
	  muteShare_internalRemoveDownloadInfo (downloadID);
	  delete downloadSemaphore;
	  delete sources;

	  if (logTimeoutChanges)
	    {
	      fclose (timeoutLogFile);
	    }

	  return MUTE_SHARE_FILE_TRANSFER_CANCELED;
	}
      else
	{

	  if (handlerResponse == 2)
	    {
	      // force de-queued

	      // remove ourself from queue
	      muteDownloadDataLock.lock ();

	      muteDownloadQueue->deleteElement (muteDownloadQueue->
						getElementIndex ((long)
								 downloadID));

	      muteNumActiveDownloads++;

	      muteDownloadDataLock.unlock ();
	    }


	  // tell handler that we are out of queue (send dequeued flag)
	  handlerResponse = inFileChunkHandler (NULL, -3, -1,
						inExtraHandlerArgument);
	}
    }





  // start out with destroyable empty strings
  char *currentSourceAddress = stringDuplicate ("");
  char *currentSourceFilePath = stringDuplicate ("");


  // search for more addresses, 10 second timeout
  sources->searchForMoreSources (10000);




  int length;
  int chunkCount;
  char *mimeType;


  int status = muteShare_internalGetFileInfo (sources,
					      &length,
					      &chunkCount,
					      &mimeType,
					      currentTimeoutMilliseconds,
					      downloadID,
					      inFileChunkHandler,
					      inExtraHandlerArgument);

  if (mimeType != NULL)
    {
      // we don't need the mime type
      delete[]mimeType;
    }

  if (status != MUTE_SHARE_FILE_INFO_FOUND)
    {
      // clean up download info
      muteShare_internalRemoveDownloadInfo (downloadID);
      delete downloadSemaphore;
      delete sources;

      if (logTimeoutChanges)
	{
	  fclose (timeoutLogFile);
	}

      muteDownloadDataLock.lock ();
      // one less active download
      muteNumActiveDownloads--;
      muteDownloadDataLock.unlock ();

      delete[]currentSourceAddress;
      delete[]currentSourceFilePath;


  if (status == MUTE_SHARE_FILE_INFO_CANCELED)
    {
      return MUTE_SHARE_FILE_TRANSFER_CANCELED;
    }
  else if (status == MUTE_SHARE_FILE_INFO_TIMEOUT)
    {
      return MUTE_SHARE_FILE_TRANSFER_FAILED;
    }
  else if (status == MUTE_SHARE_FILE_INFO_NOT_FOUND)
    {
      logMessage =
	autoSprintf
	("Aborting file get:  file info not found for file  %s : %s",
	 inVirtualAddress, inFilePath);
      AppLog::error ("fileTransfer", logMessage);
      delete[]logMessage;
    }

      return MUTE_SHARE_FILE_NOT_FOUND;

    }




  // hash the incoming file data
  QCryptographicHash hash (QCryptographicHash::Sha1);

  // if a partial file exists, pass it through the hash function
  if (inPartialFilePath != NULL)
    {
      QFile ficin (inPartialFilePath);

      if (ficin.open (QIODevice::ReadOnly))
	{
	  QByteArray buf;
	  do
	    {
	      buf = ficin.read (10240);
	      hash.addData (buf);
	    }
	  while (buf.size () == 10240);

	}

      else
	{
	  logMessage =
	    autoSprintf
	    ("Aborting file get:  failed to hash existing partial file %s",
	     inPartialFilePath);
	  AppLog::error ("fileTransfer", logMessage);
	  delete[]logMessage;


	  // clean up download info
	  muteShare_internalRemoveDownloadInfo (downloadID);
	  delete downloadSemaphore;
	  delete sources;

	  muteDownloadDataLock.lock ();
	  // one less active download
	  muteNumActiveDownloads--;
	  muteDownloadDataLock.unlock ();

	  delete[]currentSourceAddress;
	  delete[]currentSourceFilePath;

	  return MUTE_SHARE_FILE_TRANSFER_HASH_MISMATCH;
	}
    }

  RandomSource *randSource = new StdRandomSource ();


  // set the size return value here so that the handler can use it
  *outFileSizeInBytes = length;


  int startChunk = 0;

  if (inPartialFilePath != NULL && inPartialFileSizeInBytes > 0)
    {
      // skip some chunks since we already have them

      // this is not a flawless way to compute the start chunk, since
      // the file host may be using a different chunk size than us.
      // however, it should work for now, since it is fair to assume
      // that all MUTE nodes are using 16KiB chunks.
      startChunk = inPartialFileSizeInBytes / muteShareChunkSize;
    }



  delete[]currentSourceAddress;
  delete[]currentSourceFilePath;


  char chunkMissed = false;
  char canceled = false;
#define MAX_SIM_CHUNKS 10
  int handlerID[MAX_SIM_CHUNKS];
  int j;
  for (j = 0; j < MAX_SIM_CHUNKS; j++)
	  handlerID[j]=0;
  // register handlers for the chunks
  ShareFileChunkWrapper fileChunk[MAX_SIM_CHUNKS];
  for (j = 0; j < MAX_SIM_CHUNKS && (j + startChunk) < chunkCount; j++)
    {
      fileChunk[j].mChunkNumber = startChunk + j;
      fileChunk[j].mSemaphore = downloadSemaphore;
      fileChunk[j].mDataReceived = false;
      fileChunk[j].mChunkData = NULL;
      {

	// get new source

	sources->getNextSource (&currentSourceAddress,
				&currentSourceFilePath);
	fileChunk[j].mVirtualAddress = currentSourceAddress;
	fileChunk[j].mFilePath = currentSourceFilePath;

	handlerID[j] =
	  muteAddMessageHandler (muteShare_internalFileChunkHandler,
				 (void *) &fileChunk[j]);


	// send out a request message for the chunk
	char *encodedFilePath = URLUtils::hexEncode (currentSourceFilePath);

	char *message = autoSprintf ("MessageType: FileChunkRequest\n"
				     "FilePath: %s\n" "Chunknumber: %d",
				     encodedFilePath,
				     startChunk + j);

	delete[]encodedFilePath;

	if (useFreshRouteForNextChunk)
	  {
	    muteSendMessage (muteShareVirtualAddress,
			     currentSourceAddress, message, "FRESH_ROUTE");
	    useFreshRouteForNextChunk = false;
	  }
	else
	  {
	    muteSendMessage (muteShareVirtualAddress,
			     currentSourceAddress, message);
	  }
	delete[]message;
      }
    }


  while (startChunk < chunkCount && !chunkMissed && !canceled)
    {


      QTime requestStart;
      requestStart.start ();

      // wait for our handler to get the response

      int responseReceived =
	fileChunk[0].mSemaphore->tryAcquire (1, currentTimeoutMilliseconds);

      unsigned long responseTime = requestStart.elapsed ();

      if (muteShare_internalIsDownloadInterrupted (downloadID))
	{
	  // interrupted, so we didn't really receive a response from
	  // the handler
	  responseReceived = 0;
	}
      else if (responseReceived)
	{
	  // report response time to source queue
	  sources->reportBestSourceChunkTime (responseTime);
	}
      else
	{
	  // report failure
	  sources->reportBestSourceChunkTime (-1);
	}

      int numRetries = 0;


      // update our longest timeout tracker
      if (currentTimeoutMilliseconds > longestTimeoutMilliseconds)
	{
	  longestTimeoutMilliseconds = currentTimeoutMilliseconds;
	}


      // loop to perform multiple retries, as needed, for this chunk
      while (responseReceived != 1 && numRetries < maxNumRetries && !canceled)
	{
	  // no response, timed out, so retry

	  // first ask chunk handler if we should keep going
	  // pass NULL to it to indicate that we timed out on the current
	  // chunk
	  char keepGoing = inFileChunkHandler (NULL, -1, -1,
					       inExtraHandlerArgument);
	  if (!keepGoing)
	    {
	      canceled = true;
	    }
	  else
	    {
	      // not canceled, so retry


	      // should we use a freshroute?
	      // pick at random, but always freshroute on our last retry
	      char useFreshRoute = false;
	      if (randSource->getRandomFloat () <= freshRouteProbability)
		{
		  useFreshRoute = true;
		}

	      if (numRetries == maxNumRetries - 1)
		{
		  // last retry
		  // go down fighting with a FRESH_ROUTE
		  useFreshRoute = true;
		}

	      char *flags;
	      char *extraLogMessage;

	      if (useFreshRoute)
		{
		  flags = "FRESH_ROUTE";
		  extraLogMessage = "and trying a FRESH_ROUTE";

		  // set the flag, since we are deciding to use a
		  // fresh route now
		  useFreshRouteForNextChunk = true;
		}
	      else
		{
		  flags = NULL;
		  extraLogMessage = "";
		}





	      if (numRetries == maxNumRetries - 1)
		{
		  // last retry...  use a longer timeout
		  // double the longest timeout that we have ever used
		  retryTimeoutMilliseconds = 2 * longestTimeoutMilliseconds;
		}


	      logMessage =
		autoSprintf
		("Timed out, so sending retry %d with timeout of %d ms, "
		 "%s for chunk %d of " "file %s : %s", numRetries,
		 retryTimeoutMilliseconds, extraLogMessage, startChunk,
		 currentSourceAddress, currentSourceFilePath);
//              printf("%s\n",logMessage);

	      AppLog::info ("fileShare", logMessage);

	      if (logTimeoutChanges)
		{
		  fprintf (timeoutLogFile, "%s\n\n", logMessage);
		  fflush (timeoutLogFile);
		}

	      delete[]logMessage;

	      for (j = 0; j < MAX_SIM_CHUNKS; j++)
		{
		  if (fileChunk[j].mDataReceived)
		    continue;
		  // lock the file chunk to avoid access to bad data by handler
		  // as we switch sources
		  fileChunk[j].mLock.lock ();

		  // get new source
		  delete[]fileChunk[j].mVirtualAddress;
		  delete[]fileChunk[j].mFilePath;

		  sources->getNextSource (&currentSourceAddress,
					  &currentSourceFilePath);
		  fileChunk[j].mVirtualAddress = currentSourceAddress;
		  fileChunk[j].mFilePath = currentSourceFilePath;

		  fileChunk[j].mLock.unlock ();


		  // format message for new source
		  char *encodedFilePath =
		    URLUtils::hexEncode (currentSourceFilePath);

		  char *message =
		    autoSprintf ("MessageType: FileChunkRequest\n"
				 "FilePath: %s\n" "Chunknumber: %d",
				 encodedFilePath,
				 fileChunk[j].mChunkNumber);

		  delete[]encodedFilePath;

		  // use separate timing so we don't interfere with timeout
		  // code

		  // the timeout code is primarily intended for cases when
		  // we are dealing with only one source

		  // it will still function in the same way when we only have
		  // one source (setting timeout conservatively by using total
		  // time accross retries).

		  // What will it do when we have more than one source?
		  // It will set timeout too high (when one source fails and
		  // we move on to another source)
		  // After switching to the faster source, however, the
		  // timeout will be quickly readjusted.
		  // So this is okay.

		  // Don't reset the response timer here.
		  // Thus, the chunk time will include the time taken by
		  // all retries combined.  This will give us a conservative
		  // (high) estimate of the actual chunk time, which we want
		  // to set conservative timeouts.
		  // We want to *avoid* a cycle where we have overly short
		  // timeouts and get each chunk through retries, but never
		  // get good round-trip estimates during retries, since retries
		  // can make round-trip times look arbitrarily short.

		  // Unecessary retries (when a long timeout would suffice)
		  // cause undue network load.

		  // reset start time
		  // Time::getCurrentTime( &requestStartSeconds,
		  //                       &requestStartMilliseconds );

		  muteSendMessage (muteShareVirtualAddress,
				   currentSourceAddress, message, flags);

		  delete[]message;
		}
	      QTime responseStart;
	      responseStart.start ();

	      // wait again for our handler to get the response

	      responseReceived =
		fileChunk[0].mSemaphore->tryAcquire (1,
						     retryTimeoutMilliseconds);


	      responseTime = responseStart.elapsed ();

	      if (muteShare_internalIsDownloadInterrupted (downloadID))
		{
		  // interrupted, so we didn't really receive a response from
		  // the handler
		  responseReceived = 0;
		}
	      else if (responseReceived)
		{
		  // report response time to source queue
		  sources->reportBestSourceChunkTime (responseTime);
		}
	      else
		{
		  // report failure
		  sources->reportBestSourceChunkTime (-1);
		}

	      numRetries++;
	    }


	  if (!responseReceived && !canceled && numRetries == maxNumRetries)
	    {

	      // we've retried too many times

	      // in MUTE 0.4, we would simply give up an fail here

	      // instead, we can go into a kind of hibernation mode where
	      // we periodically wake up and send out a search for more
	      // sources


	      // first, reset our information so that if we actually
	      // do find sources, we will be ready to download chunks from
	      // them
	      numRetries = 0;

	      currentTimeoutMilliseconds = inTimeoutInMilliseconds;

	      retryTimeoutMilliseconds = 2 * currentTimeoutMilliseconds;

	      bestChunkTimeMilliseconds = 3600000;

	      longestTimeoutMilliseconds = currentTimeoutMilliseconds;

	      useFreshRouteForNextChunk = false;


	      // result of hibernation either found sources or canceled
	      canceled =
		!muteShare_internalHibernateForMoreSources (downloadID,
							    inFileChunkHandler,
							    inExtraHandlerArgument);
	    }


	}			// end of the retry loop



      if (canceled)
	{
	  // canceled during a retry
	}
      else if (responseReceived == 1)
	{

	  unsigned long lastChunkTimeInMilliseconds = requestStart.elapsed ();
//printf("GetFile:response received\n");

	  // Deal with timing information that comes from multiple
	  // retries as a special case.
	  // If our chunk time is longer than our current timeout,
	  // we might be over-estimating our chunk time.
	  if (lastChunkTimeInMilliseconds < currentTimeoutMilliseconds)
	    {

	      // update timeouts relative to this chunk's transfer time


	      // do we have a new best chunk time?
	      if (lastChunkTimeInMilliseconds < bestChunkTimeMilliseconds)
		{
		  bestChunkTimeMilliseconds = lastChunkTimeInMilliseconds;
		}


	      // take weighted average of current timeout with double time
	      // for our last chunk
	      // The formula:
	      // currentTimeout =
	      //    ( A * (2 * lastChunkTime) + B * currentTimeout ) / (A+B);
	      currentTimeoutMilliseconds =
		(unsigned long) ((recentChunkTimeWeight *
				  (2 * lastChunkTimeInMilliseconds)
				  +
				  currentTimeoutWeight *
				  currentTimeoutMilliseconds) /
				 (recentChunkTimeWeight +
				  currentTimeoutWeight));



	      // our timeout should always be twice as large as our
	      // most recent chunk time
	      // It decays gradually with the previous weighted formula
	      // but we always increase it immediately here if it is too low.
	      if (currentTimeoutMilliseconds <
		  2 * lastChunkTimeInMilliseconds)
		{

		  currentTimeoutMilliseconds =
		    2 * lastChunkTimeInMilliseconds;
		}



	      // log the timeout change
	      char *logMessage =
		autoSprintf ("Saw a chunk with time of %d ms, "
			     "so changing timeout to %d ms, "
			     "chunk %d of file %s : %s",
			     lastChunkTimeInMilliseconds,
			     currentTimeoutMilliseconds, startChunk,
			     currentSourceAddress, currentSourceFilePath);
	      AppLog::detail ("fileShare", logMessage);

	      if (logTimeoutChanges)
		{
		  fprintf (timeoutLogFile, "%s\n\n", logMessage);
		  fflush (timeoutLogFile);
		}

	      delete[]logMessage;
	    }
	  else
	    {
	      // don't double our last chunk time here, since this can
	      // result in a cycle of exponential timeout growth in a
	      // situation where the network is dropping a large fraction
	      // of our requests or responses

	      // any multiple that is greater than 1 (such as 1.10)
	      // could also result in exponential growth
	      // instead, simply *add* a fixed value to our last chunk
	      // timeout.  Any value will work, since it will allow
	      // our timeout to grow steadily to an appropriate value in
	      // the case where the response time has actually slowed
	      // suddenly, and it will prevent exponential growth in
	      // the case where the network is dropping a large fraction
	      // of our messages.
	      // Use 1 second here for now.
	      int addition = 1000;

	      int oldCurrentTimeout = currentTimeoutMilliseconds;

	      currentTimeoutMilliseconds =
		addition + lastChunkTimeInMilliseconds;


	      // log the chunk time
	      char *logMessage =
		autoSprintf ("Saw a chunk with time of %d ms, "
			     "but or current timeout was only %d ms, "
			     "so NOT updating timeouts the usual way, "
			     "instead adding %d ms to chunk time to get a new "
			     " timeout of %d ms, " "chunk %d of file %s : %s",
			     lastChunkTimeInMilliseconds,
			     oldCurrentTimeout, addition,
			     currentTimeoutMilliseconds,
			     startChunk, currentSourceAddress,
			     currentSourceFilePath);
	      AppLog::detail ("fileShare", logMessage);

	      if (logTimeoutChanges)
		{
		  fprintf (timeoutLogFile, "%s\n\n", logMessage);
		  fflush (timeoutLogFile);
		}

	      delete[]logMessage;
	    }


	  // retry timeout is always double our current timeout
	  retryTimeoutMilliseconds = 2 * currentTimeoutMilliseconds;


	  if (useFreshRouteForNextChunk)
	    {
	      // we were using a fresh route, which should give us
	      // the best possible chunk transfer time given the state
	      // of the network.
	      // our old bestChunkTimeMilliseconds is outdated

	      // but only update it if we weren't retrying, since
	      // retries can produce incorrect chunk time measurements,
	      // as mentioned above
	      if (numRetries == 0)
		{
		  bestChunkTimeMilliseconds = lastChunkTimeInMilliseconds;
		}

	      useFreshRouteForNextChunk = false;
	    }
	  else
	    {
	      // not using a fresh route for this chunk

	      // should we use a fresh route for the next chunk?

	      // be cautious about use of FRESH_ROUTES, since they are
	      // a burden on the network... only use one if our current
	      // chunk took 10 times as long as our best chunk.
	      if (lastChunkTimeInMilliseconds >
		  10 * bestChunkTimeMilliseconds)
		{

		  // our chunk times are much worse now than they used to
		  // be... try a fresh route to improve them

		  useFreshRouteForNextChunk = true;

		  char *logMessage =
		    autoSprintf
		    ("Time of %d ms (best time %d ms) for chunk %d of "
		     "file %s : %s, trying "
		     "a freshroute for the next chunk.",
		     lastChunkTimeInMilliseconds,
		     bestChunkTimeMilliseconds,
		     startChunk, currentSourceAddress,
		     currentSourceFilePath);
		  AppLog::info ("fileShare", logMessage);

		  if (logTimeoutChanges)
		    {
		      fprintf (timeoutLogFile, "%s\n\n", logMessage);
		      fflush (timeoutLogFile);
		    }

		  delete[]logMessage;
		}
	    }
	  int datasfound = true;
	  while (datasfound)
	    {
	      datasfound = false;
	      for (j = 0; j < MAX_SIM_CHUNKS; j++)
		{
//                      if(fileChunk[j].mDataReceived) printf("received data for chunk %d(%d desired) in handler %d\n",fileChunk[j].mChunkNumber,startChunk,j);
		  if (fileChunk[j].mDataReceived
		      && fileChunk[j].mChunkNumber == startChunk)
		    {
		      datasfound = true;
//                      printf("passing data for chunk %d in handler %d\n",startChunk,j);
		      // pass the chunk to the caller's handler
		      char keepGoing =
			inFileChunkHandler (fileChunk[j].mChunkData,
					    fileChunk[j].mLengthInBytes,
					    fileChunk[j].mChunkNumber,
					    inExtraHandlerArgument);

		      // add data to the hash
		      hash.addData ((const char *) fileChunk[j].mChunkData,
				    fileChunk[j].mLengthInBytes);
		      startChunk++;
		      delete [] fileChunk[j].mChunkData;
		      fileChunk[j].mChunkData = NULL;
		      fileChunk[j].mDataReceived = false;
		      fileChunk[j].mChunkNumber =
			startChunk + MAX_SIM_CHUNKS - 1;
		      muteRemoveMessageHandler (handlerID[j]);
		      handlerID[j]=0;
		      if (fileChunk[j].mChunkNumber < chunkCount)
			{

			  // get new source
			  delete[]fileChunk[j].mVirtualAddress;
			  delete[]fileChunk[j].mFilePath;

			  sources->getNextSource (&currentSourceAddress,
						  &currentSourceFilePath);
			  fileChunk[j].mVirtualAddress = currentSourceAddress;
			  fileChunk[j].mFilePath = currentSourceFilePath;

			  handlerID[j] =
			    muteAddMessageHandler (muteShare_internalFileChunkHandler,
			     (void *) &fileChunk[j]);


			  // send out a request message for the chunk
			  char *encodedFilePath =
			    URLUtils::hexEncode (currentSourceFilePath);

			  char *message =
			    autoSprintf ("MessageType: FileChunkRequest\n"
					 "FilePath: %s\n" "Chunknumber: %d",
					 encodedFilePath,
					 fileChunk[j].mChunkNumber);

			  delete[]encodedFilePath;

			  if (useFreshRouteForNextChunk)
			    {
			      muteSendMessage (muteShareVirtualAddress,
					       currentSourceAddress, message,
					       "FRESH_ROUTE");
			      useFreshRouteForNextChunk = false;
			    }
			  else
			    {
			      muteSendMessage (muteShareVirtualAddress,
					       currentSourceAddress, message);
			    }
			  delete[]message;
			}

		      if (!keepGoing)
			{
			  canceled = true;
			}
		    }
		}
	    }
	}
      else
	{
	  // failed completely, retries and all

	  char *logMessage =
	    autoSprintf
	    ("Giving up:  timed out (after %d retries, with a final retry "
	     "timeout of %d ms) on chunk %d " "of file %s : %s",
	     maxNumRetries,
	     retryTimeoutMilliseconds,
	     startChunk, currentSourceAddress, currentSourceFilePath);
	  AppLog::warning ("fileShare", logMessage);

	  if (logTimeoutChanges)
	    {
	      fprintf (timeoutLogFile, "%s\n\n", logMessage);
	      fflush (timeoutLogFile);
	    }

	  delete[]logMessage;

	  chunkMissed = true;

	}


      /*TODO
         if( fileChunk->mChunkData != NULL ) {
         // always delete data here (instead of only when we know we
         // have received a chunk), because:

         // 1) maybe handler was called between our timeout and
         //    our call to muteRemoveMessageHandler, or

         // 2) it could have been called between the time our semp got
         //    an external cancel signal and
         //    our call to muteRemoveMessageHandler

         delete [] fileChunk->mChunkData;
         fileChunk->mChunkData = NULL;
         }
       */
    }
  for (j = 0; j < MAX_SIM_CHUNKS; j++)
    {
      if(handlerID[j])
          muteRemoveMessageHandler (handlerID[j]);
    }

  muteShare_internalRemoveDownloadInfo (downloadID);

  delete downloadSemaphore;
  delete sources;


  if (logTimeoutChanges)
    {
      fclose (timeoutLogFile);
    }


  // compute the final hash
  char *digestHexString = strdup (hash.result ().toHex ().toUpper ().data ());


  // compare to supplied hash
  char hashMismatch = false;
  if (strcmp (digestHexString, inFileHash) != 0)
    {
      hashMismatch = true;
    }
  free (digestHexString);

  delete randSource;


  muteDownloadDataLock.lock ();
  // one less active download
  muteNumActiveDownloads--;
  muteDownloadDataLock.unlock ();


  if (chunkMissed)
    {
      return MUTE_SHARE_FILE_TRANSFER_FAILED;
    }
  else if (canceled)
    {
      return MUTE_SHARE_FILE_TRANSFER_CANCELED;
    }
  else if (hashMismatch)
    {
      // transfer succeeded (and not canceled), but hash doesn't match
      return MUTE_SHARE_FILE_TRANSFER_HASH_MISMATCH;
    }
  else
    {
      return MUTE_SHARE_FILE_TRANSFER_COMPLETE;
    }
}



void
muteShareInterruptDownload (int inDownloadID)
{
  muteDownloadDataLock.lock ();
  char found = false;
  for (int i = 0; i < muteDownloadIDs->size () && !found; i++)
    {
      if (inDownloadID == *(muteDownloadIDs->getElement (i)))
	{

	  // set interrupted flag
	  char *flag = muteDownloadInterruptedFlags->getElement (i);
	  *flag = true;

	  QSemaphore *sem = *(muteDownloadSemaphores->getElement (i));
	  sem->release ();
	  found = true;
	}
    }

  muteDownloadDataLock.unlock ();
}



int
muteShareGetSourceCount (int inDownloadID)
{
  muteDownloadDataLock.lock ();

  SourceQueue *sources =
    muteShare_internalGetDownloadSourceQueue (inDownloadID);

  int count;

  if (sources == NULL)
    {
      count = 0;
    }
  else
    {
      count = sources->getActiveSourceCount ();
    }

  muteDownloadDataLock.unlock ();

  return count;
}



char *
muteShareGetBestSource (int inDownloadID)
{

  muteDownloadDataLock.lock ();

  SourceQueue *sources =
    muteShare_internalGetDownloadSourceQueue (inDownloadID);

  char *address;

  if (sources == NULL)
    {
      address = stringDuplicate ("");
    }
  else
    {
      char *path;

      sources->getNextSource (&address, &path);

      delete[]path;
    }

  muteDownloadDataLock.unlock ();

  return address;
}



char
muteShareIsDownloadHibernating (int inDownloadID)
{
  muteDownloadDataLock.lock ();

  SourceQueue *sources =
    muteShare_internalGetDownloadSourceQueue (inDownloadID);

  char hibernating = false;

  if (sources != NULL)
    {
      hibernating = sources->isHibernating ();
    }

  muteDownloadDataLock.unlock ();

  return hibernating;
}



int
muteShareGetUploadStatus (int **outUploadIDs,
			  char ***outHostAddresses,
			  char ***outFilePaths,
			  int **outChunksInFile,
			  int **outLastChunks,
			  unsigned long **outFirstChunkTimes,
			  unsigned long **outLastChunkTimes)
{
  int *returnIDs;
  char **returnHostAddresses;
  char **returnPaths;
  int *returnLastChunks;
  int *returnChunksInFile;
  unsigned long *returnFirstChunkTimes;
  unsigned long *returnLastChunkTimes;

  muteUploadDataLock.lock ();

  int numUploads = muteUploadIDs->size ();

  returnIDs = new int[numUploads];
  returnHostAddresses = new char *[numUploads];
  returnPaths = new char *[numUploads];
  returnLastChunks = new int[numUploads];
  returnChunksInFile = new int[numUploads];
  returnFirstChunkTimes = new unsigned long[numUploads];
  returnLastChunkTimes = new unsigned long[numUploads];

  for (int i = 0; i < numUploads; i++)
    {
      returnIDs[i] = *(muteUploadIDs->getElement (i));
      returnHostAddresses[i] =
	stringDuplicate (*(muteUploadHostAddresses->getElement (i)));
      returnPaths[i] =
	stringDuplicate (*(muteUploadFilePaths->getElement (i)));

      returnLastChunks[i] = *(muteUploadLastChunksSent->getElement (i));
      returnChunksInFile[i] = *(muteUploadChunksInFiles->getElement (i));
      returnFirstChunkTimes[i] = *(muteUploadFirstChunkTimes->getElement (i));
      returnLastChunkTimes[i] = *(muteUploadLastChunkTimes->getElement (i));
    }

  muteUploadDataLock.unlock ();


  *outUploadIDs = returnIDs;
  *outHostAddresses = returnHostAddresses;
  *outFilePaths = returnPaths;
  *outLastChunks = returnLastChunks;
  *outChunksInFile = returnChunksInFile;
  *outFirstChunkTimes = returnFirstChunkTimes;
  *outLastChunkTimes = returnLastChunkTimes;

  return numUploads;
}
