/***************************************************************************
 *   Copyright (C) 2005 by Niels Sandholt Busch   *
 *   niels@diku.dk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "download.h"
#include "downloadfileevent.h"
#include "downloadstatusevent.h"
#include "downloadthread.h"
#include "kommutedefs.h"
#include "kommuteutils.h"
#include "mutelayer.h"
#include "DLListDelegate.h"
#include "xprogressbar.h"

#include <QEvent>
#include <QStandardItemModel>
#include <QStandardItem>
#include <QModelIndex>
#include <QFile>
#include <QTextStream>

extern char *DEFAULT_INCOMINGHASH_DIR;

Download::Download(QStandardItemModel *model, QString fname, unsigned long fsize, QString fhash, QString fhost)
 : QObject(0),
   name( fname ),
   size( fsize ),
   hash( fhash ),
   m_host( fhost ),
   dlThread( NULL ),
   queueState( DownloadQueue::QUEUED ),
   model( model ),
   routeQuality( -1 ),
   currentSize( 0 ),
   currentRate( 0 ),
   nbStart( 0 )
{
    int c = model->rowCount();
    model->insertRow(c);

    model->setHeaderData(c,Qt::Vertical, (quintptr)this);
    model->setData(model->index(c,NAME), fname);
    model->setData(model->index(c,HASH), fhash);
    model->setData(model->index(c,PTR), (quintptr)this);

    setProgress( 0, fsize );
    updateProgress();
    setRouteQuality( routeQuality );
    setStatus( DOWNLOADSTATUS_QUEUED );
    setHost( fhost);
    writeHash();
}

void Download::writeHash()
{
    //JMT : create file hash with contents=name of file
    char *sharePathString = MuteLayer::muteLayerGetSharingPath();
    QString fullName( sharePathString );
    delete [] sharePathString ;
    // Add the incoming download path.
    fullName += "/";
    fullName += DEFAULT_INCOMINGHASH_DIR;
    fullName += "/";
    fullName += hash;
    QFile FileHash ( fullName );
    if ( !FileHash.open( QIODevice::WriteOnly ) )
    {
	    return ;
    }
    QTextStream Stream ( & FileHash );
    Stream <<name <<endl;
    Stream <<size <<endl;
    Stream <<m_host <<endl;
    FileHash.close();
}


Download::~Download()
{
    cancel();
    if( dlThread)
    {
        dlThread->stop();
	dlThread->wait();
        delete dlThread;
        dlThread = NULL;
    }
}

void Download::customEvent( QEvent *e )
{
    int downloadStatus;

    switch( e->type() )
    {
        case EVENTTYPE_DOWNLOADFILE:
	 {
	  char * newHost= MuteLayer::getBestSource( static_cast<DownloadFileEvent*>(e)->downloadId ) ;
	   setHost( newHost );
	   delete [] newHost;
	 }
	    setSources( MuteLayer::getSourceCount( static_cast<DownloadFileEvent*>(e)->downloadId ));
	    if(currentStatus != DOWNLOADSTATUS_DOWNLOADING && blockStartSize[0]==0)
	    {
		blockStartSize[0]= static_cast<DownloadFileEvent*>(e)->currentSize ;
	    }
            setProgress( static_cast<DownloadFileEvent*>(e)->currentSize, static_cast<DownloadFileEvent*>(e)->finalSize );
            break;
        case EVENTTYPE_DOWNLOADSTATUS:
            downloadStatus = static_cast<DownloadStatusEvent*>(e)->downloadStatus;
            setStatus( downloadStatus );
            switch( downloadStatus )
            {
                // Active
                case DOWNLOADSTATUS_CONNECTING:
                    queueState = DownloadQueue::ACTIVE;
                    break;
                case DOWNLOADSTATUS_DOWNLOADING:
                    if( routeQuality < 10 )
                    {
                        routeQuality++;
                        setRouteQuality( routeQuality );
                    }
                    queueState = DownloadQueue::ACTIVE;
                    break;
                case DOWNLOADSTATUS_TIMED_OUT_WILL_RETRY:
                    if( routeQuality > 0 )
                    {
                        routeQuality--;
                        setRouteQuality( routeQuality );
                    }
                    queueState = DownloadQueue::ACTIVE;
                    break;
                case DOWNLOADSTATUS_DEQUEUED:
                    queueState = DownloadQueue::ACTIVE;
                    break;
                case DOWNLOADSTATUS_WAITING_FOR_SOURCES:
                    queueState = DownloadQueue::ACTIVE;
                    // TODO: Clear the route quality and host fields
                    break;
                case DOWNLOADSTATUS_FETCHING_INFO:
                    queueState = DownloadQueue::ACTIVE;
                    break;
                case DOWNLOADSTATUS_STARTING:
                    queueState = DownloadQueue::ACTIVE;
                    break;

                // Queued
                case DOWNLOADSTATUS_QUEUED:
                    queueState = DownloadQueue::QUEUED;
                    break;

                // Inactive
                default:
                    queueState = DownloadQueue::INACTIVE;
                    emit downloadEnded( this->host() );
                    break;
            }
            break;
        default:
            break;
    }
    updateProgress();
}

void Download::setName( QString name )
{
       model->setData(model->index(myRow(),NAME), name);
}

void Download::setHost( QString host )
{
        model->setData(model->index(myRow(),HOST), MuteLayer::getMnemonic(host) );
    m_host=host;
    writeHash();
}

void Download::setProgress( int cSize, unsigned long fSize )
{
  currentSize=cSize;
  if(fSize >0) size=fSize;
}

int Download::myRow()
{
int c = model->rowCount();
int i;
for(i=0 ; i < c ; i++)
  if(model->data(model->index(i, HASH)).toString() == hash)
	  return i;
return -1;
}

void Download::setStatus( int status )
{
int c = myRow();
    currentStatus=status;
    switch( status )
    {
        case DOWNLOADSTATUS_CONNECTING:
            model->setData(model->index(c,STATUS), tr("Connecting"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/connecting.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_DOWNLOADING:
            model->setData(model->index(c,STATUS), tr("Downloading"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/down.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_FILENOTFOUND:
	    setSources( 0);
            model->setData(model->index(c,STATUS), tr("File not found"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/notfound-corrupt.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_FILETRANSFERFAILED:
            model->setData(model->index(c,STATUS), tr("Failed"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/notfound-corrupt.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_FILETRANSFERCANCELED:
            model->setData(model->index(c,STATUS), tr("Canceled"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/cancel.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_FILETRANSFERCOMPLETE:
            model->setData(model->index(c,STATUS), tr("Complete"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/complete.png")), Qt::DecorationRole);
            model->setData(model->index(c,PROGRESS),QIcon(xProgressBar::getPixmap(QSize(160, 17), 75, false, 1)), Qt::DecorationRole );
            setRowColor(c, QString::fromUtf8("green"));
			delHash();
            break;
        case DOWNLOADSTATUS_FILETRANSFERHASHMISMATCH:
            model->setData(model->index(c,STATUS), tr("Corrupted"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/notfound-corrupt.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_QUEUED:
            model->setData(model->index(c,STATUS), tr("Queued"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/queued.png")), Qt::DecorationRole);
            break;
        case DOWNLOADSTATUS_DELAYED:
            model->setData(model->index(c,STATUS), tr("Delayed"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/delayed.png")), Qt::DecorationRole);
            break;
	case DOWNLOADSTATUS_DEQUEUED:
	case DOWNLOADSTATUS_FETCHING_INFO:
            model->setData(model->index(c,STATUS), tr("Fetching file info"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/fileinfo.png")), Qt::DecorationRole);
	    break;
	case DOWNLOADSTATUS_WAITING_FOR_SOURCES:
            model->setData(model->index(c,STATUS), tr("Waiting for sources"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/waitingforsources.png")), Qt::DecorationRole);
	    break;
	case DOWNLOADSTATUS_STARTING:
            model->setData(model->index(c,STATUS), tr("Starting"));
            model->setData(model->index(c,SYMBOL), QIcon(QString::fromUtf8(":/resources/start.png")), Qt::DecorationRole);
	    break;
        default:
            break;
    }
}

DownloadQueue::state Download::state()
{
    return queueState;
}

void Download::start()
{
    if( dlThread)
    {
        dlThread->start();
    }
    else
      dlThread = new DownloadThread( this, name, m_host, hash );
    // Start downloading
    queueState = DownloadQueue::ACTIVE;
    emit downloadStarted( this->host() );
    writeHash();
}

void Download::cancel()
{
    setStatus( DOWNLOADSTATUS_FILETRANSFERCANCELED );

    if( dlThread)
    {
        dlThread->stop();
    }

    queueState = DownloadQueue::INACTIVE;
}

QString Download::fileName()
{
    return name;
}

QString Download::host()
{
    return m_host;
}

void Download::delHash()
{
    //JMT : delete file hash
    char *sharePathString = MuteLayer::muteLayerGetSharingPath();
    QString fullName( sharePathString );
    delete [] sharePathString;

    // Add the incoming download path.
    fullName += "/";
    fullName += DEFAULT_INCOMINGHASH_DIR;
    fullName += "/";
    fullName += hash;
    QFile FileHash ( fullName );
    FileHash.remove();
}

void Download::setRouteQuality( int quality )
{
    QString str;

    if( quality < 0 )
        str = tr("Not connected");
    else if( quality <= 3 )
        str = tr("Low");
    else if( quality > 3 && quality < 7 )
        str = tr("Medium");
    else
        str = tr("High");

    model->setData(model->index(myRow(),ROUTEQUALITY), str);
}


void Download::updateProgress( )
{
    float fractionDone=0;
    if(size >0) fractionDone= (float)currentSize / (float)size;

    int percentDone = (int)( fractionDone * 100 );

    if(nbStart==0)
    {
        blockStartTime[0] = QTime::currentTime();
        blockStartSize[0] = currentSize;
        nbStart=1;
    }
    long timeDelta = blockStartTime[nbStart-1].msecsTo( QTime::currentTime() );
//printf("updateProgress: delta=%d,nb=%d,oldBlock=%d,currSize=%d\n"
//		,timeDelta,nbStart
//		,blockStartSize[0],currentSize);
    if( timeDelta > 2000 ) // 2 second has passed since last progress update
    {
	if(nbStart>=10)
	{
	    for(int i=0 ; i< 9 ; i++)
	    {
		blockStartTime[i]=blockStartTime[i+1];
		blockStartSize[i]=blockStartSize[i+1];
	    }
	    nbStart=9;
	}
	blockStartTime[nbStart]=QTime::currentTime();
	blockStartSize[nbStart++]=currentSize;
        unsigned long sizeDelta = currentSize - blockStartSize[0];
        timeDelta = blockStartTime[0].msecsTo( blockStartTime[nbStart-1]);
	if( timeDelta>0)
	{
	    currentRate = (sizeDelta*1000)/timeDelta;
//printf("updateProgress: sdelta=%d,timeD=%d\n"
//		,sizeDelta,timeDelta);
	}
	if(timeDelta>=20000 && sizeDelta==0)
		currentRate=0;
    }
    model->setData(model->index(myRow(),PROGRESS), QString::number( percentDone ));
    model->setData(model->index(myRow(),DLSPEED), QString::number( currentRate ));
    model->setData(model->index(myRow(),SIZE), QString::number(size));
    model->setData(model->index(myRow(),COMPLETED), QString::number(currentSize));
    if (currentRate >0.01)
    {
      long remaining =(size-currentSize)/currentRate;
      model->setData(model->index(myRow(),REMAINING), QString::number(remaining));
    }
    QString done = QString( "%3% " ).arg( QString::number( percentDone ) );
    model->setData(model->index(myRow(),DONE), done);
    
    int mutesharechunksize = 16384;
    int chunksinfile = size / mutesharechunksize;
    int chunksdownloaded = currentSize / mutesharechunksize;
    
    QString chunks = QString( "%1/%2" )
							.arg( QString::number( chunksdownloaded ) )
							.arg( QString::number( chunksinfile ) );
    model->setData(model->index(myRow(),CHUNKS), chunks);
}

void Download::setSources( int sources )
{
    model->setData(model->index(myRow(),SOURCES), QString::number( sources ));
}

void Download::forceDequeue()
{
    if(dlThread)
    {
      setStatus( DOWNLOADSTATUS_STARTING );
      dlThread->dequeue();
    }
    else
      start();
}

// Set the color of a row in data model
void Download::setRowColor(int c, QColor color) 
{
  unsigned int nbColumns = model->columnCount()-1;
  for(unsigned int i=0; i<nbColumns; ++i) {
    model->setData(model->index(c, i), QVariant(color), Qt::ForegroundRole);
  }
}
