/***************************************************************************
 *   Copyright (C) 2005 by Niels Sandholt Busch                            *
 *   niels@diku.dk                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "downloadthread.h"
#include "downloadfileevent.h"
#include "downloadstatusevent.h"
#include "kommutedefs.h"
#include "mutelayer.h"

#include <QApplication>
#include <QDir>
#include <QFile>
#include <QDataStream>
#include <QtDebug>

const int DEFAULT_DOWNLOADTHREAD_TIMEOUT = 30000;
const char *DEFAULT_INCOMINGDOWNLOAD_DIR = "MUTE_incoming";

DownloadThread::DownloadThread (QObject * eventReceiver, QString fname, QString fhost, QString fhash):QThread (),
saveFile (NULL),
eventReceiver (eventReceiver),
filename (fname),
host (fhost),
filehash (fhash),
finalSize (0),
currentSize (0),
running (false), forceDequeue (false), downloadId (0), partialFilePath (0)
{
  start ();
}

DownloadThread::~DownloadThread ()
{
  terminate();
  wait(500);
  if (partialFilePath)
    free (partialFilePath);
}

void
DownloadThread::run ()
{
  running = true;

  if (!validateIncomingDirectory ())
    {
      QApplication::postEvent (eventReceiver,
			       new
			       DownloadStatusEvent
			       (DOWNLOADSTATUS_FILETRANSFERCANCELED));
      return;
    }


  // Check for existing partial file ( resume )
  QString partialFile = getIncomingPath ();
  QFile file (partialFile);
  downloadId = 0;
  if (file.exists ())
    {
      QFileInfo fInfo (file);
      currentSize = fInfo.size ();
      partialFilePath = strdup(partialFile.toLocal8Bit ().data ());
      // Post progress notification to download listviewitem
      QApplication::postEvent (eventReceiver,
			       new DownloadFileEvent (downloadId, finalSize,
						      currentSize));
      //QApplication::postEvent( eventReceiver, new DownloadStatusEvent( DOWNLOADSTATUS_DOWNLOADING ) );
    }
  // Start the file transfer.
  char * sHost,*sName,*sHash;
  sHost=strdup((char *) host.toLocal8Bit ().constData ());
  sName=strdup((char *) filename.toLocal8Bit ().constData ());
  sHash=strdup((char *) filehash.toLocal8Bit ().constData ());
  while (running)
    {
      int result =
	MuteLayer::muteLayerGetFile (sHost,sName,sHash,
				     DownloadThread::muteDownloadHandler,
				     this,
				     &downloadId,
				     &finalSize,
				     DEFAULT_DOWNLOADTHREAD_TIMEOUT,
				     partialFilePath,
				     currentSize);
      int downloadStatus = result;
      QApplication::postEvent (eventReceiver,
			       new DownloadStatusEvent (downloadStatus));
      if (downloadStatus == DOWNLOADSTATUS_FILETRANSFERCOMPLETE)
	{
	  if (saveFile != NULL)
	    {
	      fclose (saveFile);
	      saveFile = NULL;
	    }
	  if (!moveCompleted ())
	    {
	      ;
	    }
	  running = false;
	}
      else
        {
	  // wait 5 minutes before retry to download
	  for(int i=0 ; i<3000 && running ; i++)
	    usleep (100000);
        }

    }
  free(sHost);
  free(sName);
  free(sHash);
  if (saveFile != NULL)
    {
      fclose (saveFile);
      saveFile = NULL;
    }

  running = false;
}

void
DownloadThread::stop ()
{
  if (running)
    {
      running = false;
//      MuteLayer::muteLayerInterruptDownload (downloadId);
    }

}


char
DownloadThread::downloadHandler (unsigned char *inChunk,
				 int inChunkLengthInBytes, int inChunkNumber)
{
// chunk==NULL and length =-2 : file is queued
// chunk==NULL and length =-3 : file is out of queue
// chunk==NULL and length =-1 : time out
// chunk != NULL and length >0 : chunk received
//
// we must return :
//      0 = cancel this download
//      1 = continue
//      2 = force dequeue (begin download)

  if(!running) return running;

  // Download timed out, will retry
  if (inChunkLengthInBytes == -1)
    {
      if (MuteLayer::isDownloadHibernating (downloadId))
	QApplication::postEvent (eventReceiver,
				 new
				 DownloadStatusEvent
				 (DOWNLOADSTATUS_WAITING_FOR_SOURCES));
      else
	QApplication::postEvent (eventReceiver,
				 new
				 DownloadStatusEvent
				 (DOWNLOADSTATUS_TIMED_OUT_WILL_RETRY));
    }

  // Download queued locally
  if (inChunkLengthInBytes == -2)
    {
      QApplication::postEvent (eventReceiver,
			       new
			       DownloadStatusEvent (DOWNLOADSTATUS_QUEUED));
    }

  // Download is active again after being queued
  if (inChunkLengthInBytes == -3)
    {
      QApplication::postEvent (eventReceiver,
			       new
			       DownloadStatusEvent (DOWNLOADSTATUS_DEQUEUED));
    }
  if (inChunk == NULL)
    {
      if (running && inChunkLengthInBytes == -2)
	return 2;
      else
	return running;
    }

  mutex.lock();

  // Validate saveFile stream.
  if (saveFile == NULL)
    {
      if (!openSaveFile ())
	{
	  mutex.unlock();
	  return false;
	}
    }

  // Save chunk to file
  if(inChunkNumber*16384 != currentSize)
    {
	printf("incorrect chunk ???\n");
	printf("curS=%d,length=%d,chunkn=%d\n",currentSize,inChunkLengthInBytes,inChunkNumber);
	fseek(saveFile,inChunkNumber*16384, SEEK_SET );
	currentSize=inChunkNumber*16384;
    }
  if (currentSize + inChunkLengthInBytes > finalSize)
    {
      fwrite (inChunk, (finalSize - currentSize), 1, saveFile);
      currentSize += (finalSize - currentSize);
    }
  else
    {
      fwrite (inChunk, inChunkLengthInBytes, 1, saveFile);
      // Keep a running total of the current size of the downloaded file.
      currentSize += inChunkLengthInBytes;
    }
  fflush (saveFile);
  mutex.unlock();

  // Post progress notification to download listviewitem
  QApplication::postEvent (eventReceiver,
			   new DownloadFileEvent (downloadId, finalSize,
						  currentSize));
  QApplication::postEvent (eventReceiver,
			   new
			   DownloadStatusEvent (DOWNLOADSTATUS_DOWNLOADING));
  if (MuteLayer::isDownloadHibernating (downloadId))
    QApplication::postEvent (eventReceiver,
			     new
			     DownloadStatusEvent
			     (DOWNLOADSTATUS_WAITING_FOR_SOURCES));

  // Activate a queued item
  if (forceDequeue)
    {
      forceDequeue = false;
      return 2;
    }
  return running;
}

/* */

char
DownloadThread::muteDownloadHandler (unsigned char *inChunk,
				     int inChunkLengthInBytes, int inChunkNumber,
				     void *inExtraParam)
{
  return static_cast <
    DownloadThread * >(inExtraParam)->downloadHandler (inChunk,
						       inChunkLengthInBytes, inChunkNumber);
}

void
DownloadThread::dequeue ()
{
  forceDequeue = true;
  // interrupt the download so that it checks for our forceDequeue signal immediately
  MuteLayer::muteLayerInterruptDownload (downloadId);
}

bool
DownloadThread::openSaveFile ()
{
  // Open this file for writing.
  saveFile = fopen (getIncomingPath ().toLocal8Bit ().constData (), "ab");
  if (saveFile)
    return true;
  else
    return false;
}

QString
DownloadThread::getFilenameWithoutPath ()
{
//     int lastSlash = filename.findRev( '/' );
  int lastSlash = filename.lastIndexOf ("/");
  return filename.right (filename.length () - lastSlash - 1);
}

QString
DownloadThread::getIncomingDirectory ()
{
  // Get the base sharing pathname.
  char *sharePathString = MuteLayer::muteLayerGetSharingPath ();
  QString fullName (sharePathString);
  delete [] sharePathString;

  // Add the incoming download path.
  fullName += "/";
  fullName += DEFAULT_INCOMINGDOWNLOAD_DIR;
  return fullName;
}

bool
DownloadThread::validateIncomingDirectory ()
{
  QDir incomingDir (getIncomingDirectory ());
  if (!incomingDir.exists ())
    {
      // Download directory does not exist, so make it.
      if (!incomingDir.mkdir (getIncomingDirectory ()))
	{
	  return false;
	}
    }
  return true;
}

QString
DownloadThread::getIncomingPath ()
{
  QString fullName = getIncomingDirectory ();
  fullName += "/";
  fullName += getFilenameWithoutPath ();
  return fullName;
}

QString
DownloadThread::getCompletedPath ()
{
  // Get the base sharing pathname.
  char *sharePathString = MuteLayer::muteLayerGetSharingPath ();
  QString fullName (sharePathString);
  delete [] sharePathString;

  // Add the filename.
  fullName += "/";
  fullName += getFilenameWithoutPath ();
  return fullName;
}

bool
DownloadThread::moveCompleted ()
{
  QDir dir;
  return dir.rename (getIncomingPath (), getCompletedPath ());
}
