/***************************************************************************
 *   Copyright (C) 2006 by Niels Sandholt Busch                            *
 *   niels@diku.dk                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
@author Andy Gebauer - feivelda@googlemail.com
*/

#include <QUrl>
//#include <QByteArray>
#include <QtDebug>

#include "webhostcachesdownload.h"
#include "defaultsettings.h"

WebHostCachesDownload::WebHostCachesDownload( QStringList oldData, QStringList downloadLocations )
{
    downloadID=0;
    listCount=0;
    // initialize http object
    http = new QHttp();
    // initialize a byte array for the buffer
    byteArray = new QByteArray();
    // initialize the read buffer from the byteArray
    buffer = new QBuffer(byteArray);
    // initialize the webhostcacheslist empty
    webHostCachesList.clear();
    // initialize the downloadList
    list = downloadLocations;
    /* initialize the semaphore object with 0 free objects
       the semaphore is needed because the settings class
       should only start when the list is downloaded or when
       this download fails*/
    /* if you don't know what a semaphore is or how it works,
       get a book about Operating System Design*/
    semaphore = new QSemaphore();
    // connect slots
    connect(http, SIGNAL(requestFinished(int, bool)),
            this, SLOT(httpRequestFinished(int, bool)));
    // prepate the standard list from the static values in defaultsettings.h
    prepareList( oldData );
}

WebHostCachesDownload::~WebHostCachesDownload()
{
    delete http;
    delete byteArray;
    delete buffer;
    delete semaphore;
}

void WebHostCachesDownload::run()
{
    QString temp;
    // when the given list is empty, use the static one
    if ( list.isEmpty() )
    {
        // the url where we get the WebHostCaches
        temp = WEBHOSTCACHESLISTDOWNLOAD;
        // make a list out of the string
        list = temp.split(LIST_SEPARATOR);
    }
    // test if the list contains list items, should now
    if ( ! list.isEmpty() )
    {
        // for every object in the list
        for ( int index = 0; index < list.size(); index++ )
        {
            // download the file at the list position index
            download( list.at(index) );            
            // request a free object. If there is none, wait for
            semaphore->acquire();
        }
    }
    /* this function is only exited when the last download call
       was successfull or exited with an error
    */
    
    #ifdef DEBUG
    qDebug() << "WebHostCaches in static default List:" << standardListCount;
    if ( listCount > 0 )
    {
        qDebug() << "WebHostCaches extra downloaded:" << listCount - standardListCount;
    }
    #endif
    
    return;
}

void WebHostCachesDownload::download(QString file)
{
    // a new QUrl object for the URL of the file given as parameter
    QUrl *url = new QUrl(file);
    // test if the url is valid
    if ( ! url->isValid() )
    {
        // if the url isn't valid, add one free object to the semaphore
        semaphore->release();
        // and exit
        return;
    }
    // if the url is valid, set the host
    http->setHost( url->host() );
    // and get the file. The ID of the transaction is stored in downloadID
    downloadID = http->get( url->path(), buffer );
    delete url;
}

void WebHostCachesDownload::prepareList( QStringList oldData )
{
    if ( !oldData.isEmpty() )
    {
        webHostCachesList = oldData;
    }
    else
    {
        // create temp string with the standard values from defaultsettings.h
        QString temp = DEFAULT_SETTINGS_STANDARDWEBHOSTCACHE;
        // split the list with the list_separator in defaultsettings.h
        webHostCachesList = temp.split(LIST_SEPARATOR);
    }
    
    // only for debugging purpose
    #ifdef DEBUG
    standardListCount = webHostCachesList.size();
    #endif
    
    return;
}

void WebHostCachesDownload::sortList()
{
    // is the list empty? If yes, exit function
    if ( webHostCachesList.isEmpty() ) return;
    // sort the list
    webHostCachesList.sort();
    // if there are empty strings, they are at the beginning
    while ( webHostCachesList.at(0).isEmpty() )
    {
        // delete it
        webHostCachesList.removeAt(0);
        // exit the while loop when we deleted the last entry, even if it was empty
        if ( webHostCachesList.isEmpty() ) break;
    }
    // our index
    int index = 0;
    // if the point we are still not at the second item from the end
    while ( index < webHostCachesList.size() - 1 )
    {
        /* if the string at index and the next are equal, delete the second
           we don't need a complex algorithem because we sorted the list before
           so all equal strings are behind each other*/
        if ( webHostCachesList.at(index) == webHostCachesList.at(index+1) )
        {
            // delete the double value
            webHostCachesList.removeAt(index+1);
        }
        else
        {
            // go to the next position
            index++;
        }
    }
    
    // only for debugging purpose
    #ifdef DEBUG
    listCount = webHostCachesList.size();
    #endif
}

void WebHostCachesDownload::parseBuffer()
{
    // create temp string from the buffer
    QString temp = byteArray->constData();
    // append the splitted temp string to the webHostCachesList
    webHostCachesList += temp.split('\n');
    // add one free object to the semaphore
    semaphore->release();
}

QStringList WebHostCachesDownload::getWebHostCaches()
{
    // sort the list and delete double values
    sortList();
    // return the webHostCachesList
    return webHostCachesList;
}

void WebHostCachesDownload::httpRequestFinished( int ID, bool error )
{
    // if the ID isn't the downloadID from our job, then this isn't the
    // get request, so finish here. The file isn't here
    if ( ID != downloadID )
    {
        // exit the function
        return;
    }
    // if an error occured
    if ( error )
    {
        #ifdef DEBUG
        qDebug() << http->errorString();
        #endif
        // add one free object to the semaphore
        semaphore->release();
        // exit the function
        return;
    }
    // parse the buffer
    parseBuffer();
}
